import tensorflow as tf

from detext.model.bert import modeling


def create_bert_model(bert_config_path,
                      is_training,
                      input_ids,
                      bert_checkpoint_path=None,
                      input_mask=None,
                      token_type_ids=None,
                      use_one_hot_embeddings=False
                      ):
    """
    Create a Bert model instance, with the option to initialize the params from a pretrained checkpoint.
    :param bert_config_path: config path of Bert model.
    :param bert_checkpoint_path: checkpoint path for pretrained model.
    :param is_training: true for training model, false for eval model. Controls whether dropout will be applied.
    :param input_ids: int32 Tensor of shape [batch_size, seq_length].
    :param input_mask: (optional) int32 Tensor of shape [batch_size, seq_length].
    :param token_type_ids: (optional) int32 Tensor of shape [batch_size, seq_length]. token_type_ids: (optional) int32
        Tensor of shape [batch_size, seq_length].
    :param use_one_hot_embeddings: (optional) bool. Whether to use one-hot word embeddings or tf.embedding_lookup()
        for the word embeddings. On the TPU, it is much faster if this is True, on the CPU or GPU, it is faster if this
        is False.
    :return: A Bert model instance.
    """
    if not bert_config_path:
        raise ValueError("A BERT config file must be given.")
    bert_config = modeling.BertConfig.from_json_file(bert_config_path)

    bert_model = modeling.BertModel(
        config=bert_config,
        is_training=is_training,
        input_ids=input_ids,
        input_mask=input_mask,
        token_type_ids=token_type_ids,
        use_one_hot_embeddings=use_one_hot_embeddings,
    )

    if bert_checkpoint_path:
        init_from_checkpoint(bert_checkpoint_path)

    return bert_model


def init_from_checkpoint(bert_checkpoint_path):
    """
    Given a checkpoint path of a pretrained Bert model, initialize the variables.
    :param bert_checkpoint_path: pretrained model checkpoint path.
    :return:
    """
    if not bert_checkpoint_path:
        raise ValueError("A BERT checkpoint path must be given to initialize the variables.")
    tvars = tf.trainable_variables()
    assignment_map, initialized_variable_names = modeling.get_assignment_map_from_checkpoint(tvars,
                                                                                             bert_checkpoint_path)
    tf.train.init_from_checkpoint(bert_checkpoint_path, assignment_map)
    print('%d variables are initialized from bert pretrained model' % len(initialized_variable_names))



import tensorflow as tf
import tensorflow_ranking as tfr


def compute_ndcg_tfr(scores, labels, features, topk):
    """Computes NDCG using tf ranking"""
    metric_fn = tfr.metrics.make_ranking_metric_fn(tfr.metrics.RankingMetricKey.NDCG, topn=topk)
    return metric_fn(labels, scores, features)


def compute_mrr_tfr(scores, labels, features):
    """Computes MRR using tf ranking

    NOTE: There is a bug in tf-rank that's causing topk to have no effect, so this function is computing MRR for
        whole list WHATEVER topk is.
    """
    metric_fn = tfr.metrics.make_ranking_metric_fn(tfr.metrics.RankingMetricKey.MRR)
    return metric_fn(labels, scores, features)


def compute_precision_tfr(scores, labels, features, topk):
    """Computes precision using tf ranking"""
    metric_fn = tfr.metrics.make_ranking_metric_fn(tfr.metrics.RankingMetricKey.PRECISION, topn=topk)
    return metric_fn(labels, scores, features)


def compute_dcg(scores, labels, group_size, topk):
    """
    Compute the dcg score.
    """
    max_group_size = tf.reduce_max(group_size)
    batch_size = tf.shape(scores)[0]
    topk = tf.minimum(max_group_size, topk)
    # The padded docs should have lowest scores
    mask = tf.sequence_mask(group_size, maxlen=tf.shape(scores)[1], dtype=tf.float32)
    scores2 = (scores - tf.reduce_min(scores)) * mask - (1 - mask)

    def get_index(cols):
        rows = tf.reshape(tf.range(batch_size * topk) // topk, [batch_size, topk])
        index = tf.stack([rows, cols], axis=2)
        return index

    # dcg
    _, cols = tf.nn.top_k(scores2, k=topk)
    actual_index = get_index(cols)
    actual_score = tf.gather_nd(labels, actual_index)
    # for one query, it is possible that topk is larger than group_size.  The actual_mask makes sure that the padded
    # document get a score of 0 in actual_score.
    actual_mask = tf.gather_nd(mask, actual_index)
    dcg = actual_score * actual_mask / tf.log(tf.cast(tf.range(2, topk + 2), dtype=tf.float32)) * tf.log(2.0)
    dcg = tf.reduce_sum(dcg, axis=-1)
    return dcg


def compute_mrr(scores, labels, group_size, topk):
    """
    Computes MRR score

    :param scores: Tensor Predicted scores. Shape=[batch_size, max_group_size]
    :param labels: Tensor Labels. Shape=[batch_size, max_group_size]
    :param group_size: Tensor Number of documents in each sample. Shape=[batch_size]
    :param topk: int Number of samples to select in MRR computation
    :return: Metric of average MRR
    """
    max_group_size = tf.reduce_max(group_size)
    batch_size = tf.shape(scores)[0]
    topk = tf.minimum(max_group_size, topk)
    # The padded docs should have lowest scores
    mask = tf.sequence_mask(group_size, maxlen=tf.shape(scores)[1], dtype=tf.float32)
    scores2 = (scores - tf.reduce_min(scores)) * mask - (1 - mask)

    def get_index(cols):
        rows = tf.reshape(tf.range(batch_size * topk) // topk, [batch_size, topk])
        index = tf.stack([rows, cols], axis=2)
        return index

    _, cols = tf.nn.top_k(scores2, k=topk)
    actual_index = get_index(cols)
    actual_score = tf.gather_nd(labels, actual_index)
    # for one query, it is possible that topk is larger than group_size.  The actual_mask makes sure that the padded
    # document get a score of 0 in actual_score.
    actual_mask = tf.gather_nd(mask, actual_index)
    rr = actual_score * actual_mask / tf.cast(tf.range(1, topk + 1), dtype=tf.float32)
    rr = tf.reduce_max(rr, axis=-1)
    return tf.metrics.mean(rr)


def compute_ndcg(scores, labels, group_size, topk):
    """
    Computes ndcg score. Caution: it's the traditional formula instead of the one with numerator as power(2, relevance)

    :param scores: Tensor Predicted scores. Shape=[batch_size, max_group_size]
    :param labels: Tensor Labels. Shape=[batch_size, max_group_size]
    :param group_size: Tensor Number of documents in each sample. Shape=[batch_size]
    :param topk: int Number of samples to select in ndcg computation
    :return: Metric of average ndcg
    """
    idcg = compute_dcg(labels, labels, group_size, topk=topk)
    dcg = compute_dcg(scores, labels, group_size, topk=topk)

    ndcg = tf.div_no_nan(dcg, idcg)
    return tf.metrics.mean(ndcg)


def compute_preat1(scores, labels):
    """
    Compute precision@1, i.e., whether the top1 result has the highest label.
    It should be only applied to one-hot labels.
    :param scores: Tensor Predicted scores. Shape=[batch_size, max_group_size]
    :param labels: Tensor Labels. Shape=[batch_size, max_group_size]
    :return: Metric of average precision@1
    """
    # prediction
    predict = tf.argmax(scores, axis=-1, output_type=tf.int32)
    pred_onehot = tf.one_hot(predict, tf.shape(scores)[-1])
    # gold standard
    max_value = tf.tile(tf.reduce_max(labels, axis=-1, keepdims=True), [1, tf.shape(labels)[-1]])
    is_max_value = tf.cast(tf.equal(labels, max_value), dtype=tf.float32)
    # compute preat1
    preat1 = tf.reduce_sum(pred_onehot * is_max_value, axis=-1)
    return tf.metrics.mean(preat1)


def compute_rank(scores, group_size):
    """
    compute each element's rank
    """
    num_cols_scores = tf.shape(scores)[1]
    # Make sure the padded docs have lowest scores
    mask = tf.sequence_mask(group_size, maxlen=num_cols_scores, dtype=tf.float32)
    scores2 = (scores - tf.reduce_min(scores)) * mask - (1 - mask)
    # construct rank matrix using tf.nn.top_k
    _, col_idx = tf.nn.top_k(scores2, k=num_cols_scores)
    col_idx = tf.reshape(col_idx, shape=[-1])
    row_idx = tf.range(tf.size(scores)) // num_cols_scores
    idx = tf.concat([tf.expand_dims(row_idx, -1), tf.expand_dims(col_idx, -1)], axis=1)
    val = tf.cast(tf.range(tf.size(scores)) % num_cols_scores + 1, dtype=tf.float32)
    rank_mat = tf.scatter_nd(idx, val, shape=tf.shape(scores))
    return rank_mat


def compute_auc(scores, labels):
    """Computes AUC score given scores and labels.

    :param scores: Tensor Predicted scores. Shape=[batch_size, max_group_size].
        max_group_size should be 1
    :param labels: Tensor Labels. Shape=[batch_size, max_group_size]
        max_group_size should be 1
    """
    labels = tf.reshape(labels, shape=[tf.shape(labels)[0]])
    scores = tf.reshape(scores, shape=[tf.shape(scores)[0]])
    prob = tf.sigmoid(scores)
    return tf.metrics.auc(labels, prob, num_thresholds=2000)


def compute_confusion_matrix(scores, labels, num_classes):
    """
    Compute the confusion matrix for multi-class classifications.
    :param scores: Tensor Predicted scores. Shape=[batch_size, num_classes]
    :param labels: Tensor Labels. Shape=[batch_size, max_group_size]
        max_group_size should be 1
    :param num_classes: Number of classes for multi-class classification.
    :return: THe confusion matrix metric in the format of (metric_value, update_op) tuple.
    """
    labels = tf.squeeze(labels, -1)
    probabilities = tf.nn.softmax(scores)
    predicted_indices = tf.argmax(probabilities, 1)

    con_matrix = tf.confusion_matrix(labels=labels, predictions=predicted_indices,
                                     num_classes=num_classes)

    con_matrix_sum = tf.Variable(tf.zeros(shape=(num_classes, num_classes), dtype=tf.int32),
                                 trainable=False,
                                 name="confusion_matrix_result",
                                 collections=[tf.GraphKeys.LOCAL_VARIABLES])
    update_op = tf.assign_add(con_matrix_sum, con_matrix)
    return tf.convert_to_tensor(con_matrix_sum), update_op


def compute_accuracy(scores, labels):
    """
    Compute the accuracy given the scores and labels.
    :param scores: Tensor Predicted scores. Shape=[batch_size, num_classes]
    :param labels: Tensor Labels. Shape=[batch_size, max_group_size]
        max_group_size should be 1
    :return: The accuracy (% of predicted_indices matches labels)
    """
    labels = tf.squeeze(labels, -1)
    probabilities = tf.nn.softmax(scores)
    predicted_indices = tf.argmax(probabilities, 1)
    return tf.metrics.accuracy(labels, predicted_indices)


"""
A lot of non-tensorflow implementation.  This is mainly used for testing.
"""
import argparse
import math

import numpy as np
import tensorflow as tf


def neg_log_sigmoid(x):
    """Negative log sigmoid"""
    return -math.log(1 / (1 + math.exp(-x)))


def compute_dcg_power2(y_score, y_true, topk):
    """Computes dcg with relevance score power(2, y_true)"""
    order = np.argsort(y_score)[::-1]
    y_true = np.take(y_true, order[:topk])
    gain = 2 ** y_true - 1
    discounts = np.log2(np.arange(len(y_true)) + 2)
    return np.sum(gain / discounts)


def compute_ndcg_power2(y_score, y_true, topk):
    """Computes idcg with relevance score power(2, y_true)"""
    dcg = compute_dcg_power2(y_score, y_true, topk)
    print(dcg)
    idcg = compute_dcg_power2(y_true, y_true, topk)
    print(idcg)
    return dcg / idcg if idcg > 0 else 0


def get_lambda_loss(scores, labels, group_size, topk=None):
    """A non tf implementation of lambda loss with ndcg"""
    batch_size = len(scores)
    max_group_size_scores = len(scores[0])
    loss = []
    for i in range(batch_size):
        loss.append([])
        for j in range(max_group_size_scores):
            loss[i].append([])
            for k in range(max_group_size_scores):
                if j < group_size[i] and k < group_size[i] and labels[i][j] > labels[i][k]:
                    loss[i][j].append(neg_log_sigmoid(scores[i][j] - scores[i][k]))
                    if topk:  # lambda rank
                        ndcg1 = get_ndcg(scores[i][:group_size[i]], labels[i][:group_size[i]], topk)
                        new_score = list(scores[i])
                        new_score[j] = scores[i][k]
                        new_score[k] = scores[i][j]
                        ndcg2 = get_ndcg(new_score[:group_size[i]], labels[i][:group_size[i]], topk)
                        ndcg_diff = abs(ndcg1 - ndcg2)
                        loss[i][j][k] *= ndcg_diff
                else:
                    loss[i][j].append(0)
    return loss


def get_ndcg(scores, labels, topk):
    """compute ndcg"""
    # idcg
    idcg, k = 0, 0
    for i, s in enumerate(sorted(labels, reverse=True)):
        idcg += s / math.log(2 + i, 2)
        k += 1
        if k >= topk:
            break
    # dcg
    dcg, k = 0, 0
    for i, (j, _) in enumerate(sorted(enumerate(scores), key=lambda x: x[1], reverse=True)):
        dcg += labels[j] / math.log(2 + i, 2)
        k += 1
        if k >= topk:
            break
    return dcg / idcg


def get_softmax_loss(scores, labels, group_size):
    loss = []
    for score, label, group_s in zip(scores, labels, group_size):
        # find the index with largest value in label
        sm_score = []
        for i in range(len(score)):
            if i >= group_s:
                exp_score = 0
            else:
                exp_score = math.exp(score[i])
            sm_score.append(exp_score)
        sm_score_sum = sum(sm_score)

        for i in range(min(len(score), group_s)):
            sm_score[i] = -math.log(sm_score[i] / sm_score_sum) * label[i]
        loss.append(sm_score)
    return loss


def compute_ndcg_elem_ps_parity_check(result_list, label_index, top_n):
    dcg = 0
    for i, result in enumerate(result_list):
        if i >= top_n:
            break
        index = i + 1
        score = float(result[label_index])
        dcg = dcg + score / math.log(index + 1, 2)
    return dcg


def compute_ndcg_ps_parity_check_original_api(input_list, n):
    # compute dcg
    # dcg = \sum_i {label_i / log_2(pos_i + 1)}, i ordered by prediction score
    result_order_by_score = sorted(input_list, key=lambda f: f[2], reverse=True)
    dcg = compute_ndcg_elem_ps_parity_check(result_order_by_score, 1, n)

    # compute idcg
    # idcg = \sum_i {label_i / log_2(pos_i + 1)}, i ordered by relevance score (click)
    result_order_by_click = sorted(input_list, key=lambda f: f[1], reverse=True)
    idcg = compute_ndcg_elem_ps_parity_check(result_order_by_click, 1, n)

    # calculate ndcg
    ndcg = 0.0
    if idcg > 0.0:
        ndcg = dcg / idcg
    return ndcg


def compute_ndcg_ps_parity_check(scores, labels, topk=10):
    """
    adapter for pps ndcg calculating util
    :param scores: raw scores
    :param labels: actual labels, ordered numerically
    :param topk: default is 10
    :return: ndcg score as float
    """
    return compute_ndcg_ps_parity_check_original_api(list(zip(scores, labels, scores)), topk)


def add_arguments(parser):
    """Build ArgumentParser."""

    # network
    parser.add_argument("--model_dir", type=str, help="exported model dir")
    parser.add_argument("--test_files", type=str, help="test files to be evaluated on")
    parser.add_argument("--doc_text_fields", type=str, help="document text fields")
    parser.add_argument("--usr_text_fields", type=str, help="user text fields")
    parser.add_argument("--num_wide", type=int, help="number of wide features per doc")
    parser.add_argument("--output_dir", type=str, help="result file written in output dir")


def get_params(argv):
    """
    Get hyper-parameters.
    """
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    hparams, unknown_params = parser.parse_known_args(argv)

    # Print all hyper-parameters
    for k, v in sorted(vars(hparams).items()):
        print('--' + k + '=' + str(v))
    return hparams


def create_sample_tfrecord(out_file):
    """Creates sample tfrecord to out_file"""
    print("Composing fake tfrecord to file {}".format(out_file))
    with tf.io.TFRecordWriter(out_file) as writer:
        with tf.Graph().as_default(), tf.compat.v1.Session():
            # Example 1
            features = {
                "label": _float_feature(
                    [1., 0., 0., 0.],
                ),
                "query": _bytes_feature([b"hello"]),
                "wide_ftrs": _float_feature([23.0, 14.0, 44.0, -1.0, 22.0, 19.0, 22.0, 19.0]),
                "doc_title": _bytes_feature([b"document title 1", b"title 2 ?", b"doc title 3 ?", b"doc title 4 ?"]),
                "usr_headline": _bytes_feature([b"user headline"]),
                "wide_ftrs_sp_idx": _int64_feature([1, 0, 2, 8]),
                "wide_ftrs_sp_val": _float_feature([1.0, 0.0, 7.0, 12.0])
            }
            example_proto = tf.train.Example(features=tf.train.Features(feature=features))
            writer.write(example_proto.SerializeToString())

            # Example 2
            features = {
                "label": _float_feature(
                    [0., 0., 1.],
                ),
                "query": _bytes_feature([b"hello"]),
                "wide_ftrs": _float_feature([23.0, 14.0, 44.0, -1.0, 22.0, 19.0]),
                "doc_title": _bytes_feature([b"document title 1 linkedin", b"doc title 2 ?", b"doc title 3 ?"]),
                "usr_headline": _bytes_feature([b"user headline ?"]),
                "wide_ftrs_sp_idx": _int64_feature([8, 0, 2]),
                "wide_ftrs_sp_val": _float_feature([1.0, 0.0, 7.0])
            }
            example_proto = tf.train.Example(features=tf.train.Features(feature=features))
            writer.write(example_proto.SerializeToString())


def _bytes_feature(value):
    """Returns a bytes_list feature"""
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy()  # BytesList won't unpack a string from an EagerTensor.
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=value))


def _float_feature(value):
    """Returns a float_list feature"""
    return tf.train.Feature(float_list=tf.train.FloatList(value=value))


def _int64_feature(value):
    """Returns an int64_list feature"""
    return tf.train.Feature(int64_list=tf.train.Int64List(value=value))



import tensorflow as tf

from detext.utils.model_utils import create_rnn_cell, init_word_embedding


class LstmLmModel():
    def __init__(self,
                 query,
                 doc_fields,
                 usr_fields,
                 hparams,
                 mode):
        """
        Applies LSTM (language modeling version) to convert text to a fix length embedding

        :param query: Tensor(dtype=tf.int)  Shape=[batch_size, query_length]
        :param doc_fields: list(Tensor(dtype=int))/Tensor  A list of document fields. Each has shape=
            [batch_size, max_group_size, doc_field_length]. For online scoring, these fields may be precomputed and
            input as Tensor
        :param usr_fields: list(Tensor(dtype=int))/Tensor  A list of user fields. Each has shape=
            [batch_size, usr_field_length]. For online scoring, these fields may be precomputed and
            input as Tensor
        :param hparams: HParams
        :param mode: TRAIN/EVAL/INFER
        """
        self._query = query
        self._doc_fields = doc_fields
        self._usr_fields = usr_fields
        self._hparams = hparams
        self._mode = mode
        self.text_ftr_size = 1

        # Initialize embedding
        self.embedding = init_word_embedding(self._hparams, self._mode)
        # NOTE: when normalized_lm is True, this embedding is not used because normalized_lm use a different
        #   implementation (dense layer) for embedding interaction
        if hparams.normalized_lm is False:
            self.nextw_embedding = init_word_embedding(self._hparams, self._mode, name_prefix="nextw")

        with tf.variable_scope("lstm_lm", dtype=tf.float32):
            # RNN cell setup
            self.cell_dtype = tf.float32
            self.cell = create_rnn_cell(
                unit_type=hparams.unit_type,
                num_units=hparams.num_units,
                num_layers=hparams.num_layers,
                num_residual_layers=hparams.num_residual_layers,
                forget_bias=hparams.forget_bias,
                dropout=hparams.rnn_dropout,
                mode=tf.contrib.learn.ModeKeys.TRAIN,
            )
            self.length_bias = tf.get_variable("length_bias", shape=[1], dtype=tf.float32)
            if hparams.bidirectional:
                self.cell_fw = create_rnn_cell(
                    unit_type=hparams.unit_type,
                    num_units=hparams.num_units / 2,
                    num_layers=hparams.num_layers,
                    num_residual_layers=hparams.num_residual_layers,
                    forget_bias=hparams.forget_bias,
                    dropout=hparams.rnn_dropout,
                    mode=tf.contrib.learn.ModeKeys.TRAIN,
                )
                self.cell_bw = create_rnn_cell(
                    unit_type=hparams.unit_type,
                    num_units=hparams.num_units / 2,
                    num_layers=hparams.num_layers,
                    num_residual_layers=hparams.num_residual_layers,
                    forget_bias=hparams.forget_bias,
                    dropout=hparams.rnn_dropout,
                    mode=tf.contrib.learn.ModeKeys.TRAIN,
                )

            # Apply LSTM on text fields
            self.query_ftrs = self.apply_lstm_on_query() if query is not None else None
            self.usr_ftrs = self.apply_lstm_on_usr() if usr_fields is not None else None
            self.doc_ftrs = self.apply_lstm_on_doc()

    def apply_lstm_on_query(self):
        """
        Applies LSTM on query

        :return Tensor  Query features. Shape=[batch_size, 1]
        """
        return self.apply_lstm_on_text(self._query)

    def apply_lstm_on_usr(self):
        """
        Applies LSTM on user fields

        :return Tensor  User features. Shape=[batch_size, num_usr_fields, 1]
        """
        if type(self._usr_fields) is not list:
            return self._usr_fields

        usr_ftrs = []
        for i, usr_field in enumerate(self._usr_fields):
            usr_field_ftrs = self.apply_lstm_on_text(usr_field)
            usr_ftrs.append(usr_field_ftrs)
        usr_ftrs = tf.stack(usr_ftrs, axis=1)  # shape=[batch_size, num_usr_fields, 1]
        return usr_ftrs

    def apply_lstm_on_doc(self):
        """Applies LSTM on documents

        :return Tensor  Document features. Shape=[batch_size, max_group_size, num_doc_fields, 1]
        """

        # doc_fields should be the doc embeddings
        if type(self._doc_fields) is not list:
            return self._doc_fields

        doc_ftrs = []
        for i, doc_field in enumerate(self._doc_fields):
            doc_shape = tf.shape(doc_field)  # Shape=[batch_size, group_size, sent_length]
            doc_field = tf.reshape(doc_field, shape=[doc_shape[0] * doc_shape[1], doc_shape[2]])

            # Apply LSTM
            doc_field_ftrs = self.apply_lstm_on_text(doc_field)

            # Restore batch_size and group_size
            doc_field_ftrs = tf.reshape(doc_field_ftrs, shape=[doc_shape[0], doc_shape[1], self.text_ftr_size])
            doc_ftrs.append(doc_field_ftrs)
        doc_ftrs = tf.stack(doc_ftrs, axis=2)  # Shape=[batch_size, max_group_size, num_doc_fields, 1]
        return doc_ftrs

    def apply_lstm_on_text(self, text):
        """
        Applies LSTM (language modeling) on text

        :param text: Tensor Shape=[batch_size, seq_len]
        :return Tensor Shape=[batch_size, 1]
        """
        hparams = self._hparams

        # Shape=[batch_size]
        seq_len = tf.reduce_sum(tf.cast(tf.not_equal(text, hparams.pad_id), dtype=tf.int32), axis=-1)
        max_seq_len = tf.shape(text)[1]

        input_seq = text[:, :-1]  # [batch_size, seq_len - 1]
        input_emb = tf.nn.embedding_lookup(self.embedding, input_seq)  # [batch_size, seq_len - 1, num_units]
        output_seq = text[:, 1:]  # [batch_size, seq_len - 1]

        # [batch_size, seq_len - 1, num_units]
        seq_outputs, _ = tf.nn.dynamic_rnn(
            self.cell,
            input_emb,
            dtype=self.cell_dtype,
            sequence_length=seq_len - 1,
            time_major=False,
            swap_memory=True)

        # Whether to use bidirectional RNN
        if hparams.bidirectional:
            seq_outputs, _ = tf.nn.bidirectional_dynamic_rnn(
                self.cell_fw,
                self.cell_bw,
                input_emb,
                dtype=self.cell_dtype,
                sequence_length=seq_len - 1,
                time_major=False,
                swap_memory=True)
            seq_outputs = tf.concat(seq_outputs, axis=2)

        # Whether to use normalized language model
        if not hparams.normalized_lm:
            output_emb = tf.nn.embedding_lookup(self.nextw_embedding, output_seq)  # [batch_size, seq_len - 1, num_units]
            # If do reduce_sum for the result, it will approximate normalized log_p(w_i+1|w_i, w_i-1, ..., w_0)
            word_scores = seq_outputs * output_emb

            # Approximate normalized log_p(word sequence)
            word_scores = tf.reduce_sum(word_scores, axis=-1)  # [batch_size, max_seq_len - 1]
            # Use bias to approximate sum(dot(curr_embedding, all_embedding))
            word_scores += self.length_bias
        else:
            # Compute h_i*w_i+1 - log_sum_k_softmax(h_i*w_k)
            seq_outputs = tf.layers.dense(
                seq_outputs, units=hparams.vocab_size, use_bias=False, name="dot_prod_nextw",
                reuse=tf.AUTO_REUSE)  # [batch_size, seq_len - 1, vocab_size]
            word_scores = tf.nn.sparse_softmax_cross_entropy_with_logits(
                labels=output_seq, logits=seq_outputs)

        # Mask out padding scores
        mask = tf.sequence_mask(seq_len - 1, max_seq_len - 1, dtype=tf.float32)
        word_scores *= mask
        word_scores = tf.reduce_sum(word_scores, axis=-1, keepdims=True)  # [batch_size, 1]
        return word_scores



"""
softmax loss.
"""

import tensorflow as tf


def compute_softmax_loss(scores, labels, group_size):
    """
    It computes the sum of negative log softmax loss:
    -sum_i lables_i * log(exp(scores_i) / (exp(scores_1) + ... + exp(scores_n)))
    """
    # mask the padded documents
    mask = tf.sequence_mask(group_size, maxlen=tf.shape(scores)[1], dtype=tf.float32)
    # softmax loss
    loss = mask * labels * (-scores + tf.expand_dims(compute_logsumexp_mask(scores, mask), axis=1))
    return loss


def compute_sigmoid_cross_entropy_loss(scores, labels, group_size):
    """ Compute loss for pointwise ranking

    :param scores: Tensor  Shape=[batch_size, max_group_size]
    :param labels: Tensor  Shape=[batch_size, max_group_size]
    :param group_size: Tensor  Shape=[batch_size]
    :return: Tensor  Shape=[batch_size, max_group_size]
   """
    mask = tf.sequence_mask(group_size, maxlen=tf.shape(scores)[1], dtype=tf.float32)
    loss = mask * tf.nn.sigmoid_cross_entropy_with_logits(labels=labels, logits=scores)
    return tf.reduce_sum(loss, axis=-1)


def compute_logsumexp_mask(scores, mask):
    """
    Compute logsumexp with mask for a batch.
    Current tf.lossumexp does not support mask.
    """
    # Change padding scores to be the score of the first element in the score array. I.e., if
    #   original_scores = [[1., -1., PAD_SCORE], [2., -2., PAD_SCORE]], result_scores will be
    #   [[1., -1., 1.], [2., -2., 2]]. This is to make sure the maximum value is chosen from
    #   VALID scores
    scores = tf.where(tf.cast(mask, dtype=tf.bool), scores,
                      tf.ones_like(scores) * tf.expand_dims(scores[:, 0], axis=-1))
    max_scores = tf.reduce_max(scores, axis=-1, keepdims=True)
    scores = scores - max_scores
    exp_score_withmask = tf.exp(scores) * tf.cast(mask, dtype=tf.float32)
    logsumexp = tf.log(tf.reduce_sum(exp_score_withmask, axis=-1)) + tf.squeeze(max_scores, -1)
    return logsumexp


def compute_regularization_penalty(hparams):
    """ Returns the regularization penalty specified in hparams """
    penalty = 0
    if hparams.l1 is not None and hparams.l1 != 0:
        regularizer = tf.contrib.layers.l1_regularizer(scale=hparams.l1)
        penalty += compute_penalty_given_regularizer(regularizer)
    if hparams.l2 is not None and hparams.l2 != 0:  # L2 loss in TensorFlow is computed as 1/2*sum(weights**2)
        regularizer = tf.contrib.layers.l2_regularizer(scale=hparams.l2)
        penalty += compute_penalty_given_regularizer(regularizer)
    return penalty


def compute_penalty_given_regularizer(regularizer):
    """ Returns penalty given regularizer

    This function applies to all trainable variables
    """
    weights = tf.trainable_variables()
    regularization_penalty = tf.contrib.layers.apply_regularization(regularizer, weights)
    return regularization_penalty



import tensorflow as tf
import tensorflow_ranking as tfr
from os.path import join as path_join

from detext.model.deep_match import DeepMatch
from detext.model.lambdarank import LambdaRank
from detext.train import metrics
from detext.train import optimization, train_helper
from detext.train.loss import compute_softmax_loss, compute_sigmoid_cross_entropy_loss, \
    compute_regularization_penalty
from detext.utils import vocab_utils, executor_utils
from detext.utils.best_checkpoint_copier import BestCheckpointCopier


def train(hparams, input_fn):
    """
    Main function for train/evaluate DeText ranking model
    :param hparams: hparams
    :param input_fn: input function to create train/eval specs
    :return:
    """
    if hparams.use_horovod is True:
        import horovod.tensorflow as hvd
    else:
        hvd = None
    train_strategy = tf.contrib.distribute.ParameterServerStrategy()
    estimator = get_estimator(hparams, strategy=train_strategy)

    # Set model export config for evaluator or primary worker of horovod
    exporter_list = None
    if not hvd or hvd.rank() == 0:
        best_model_name = 'best_' + hparams.pmetric
        # Exporter to save best (in terms of pmetric) checkpoint in the folder [best_model_name],
        # and export to savedmodel for prediction.
        best_checkpoint_exporter = BestCheckpointCopier(
            name=best_model_name,
            serving_input_receiver_fn=lambda: serving_input_fn(hparams),
            checkpoints_to_keep=1,  # keeping the best checkpoint
            exports_to_keep=1,  # keeping the best savedmodel
            pmetric=f'metric/{hparams.pmetric}',
            compare_fn=lambda x, y: x.score > y.score,  # larger metric better
            sort_reverse=True)
        exporter_list = [best_checkpoint_exporter]

    # Handle sync distributed training case via use_horovod

        # Horovod: BroadcastGlobalVariablesHook broadcasts initial variable states from
        # rank 0 to all other processes. This is necessary to ensure consistent
        # initialization of all workers when training is started with random weights or
        # restored from a checkpoint.
    bcast_hook = [hvd.BroadcastGlobalVariablesHook(0)] if hvd else None

    def input_fn_common(pattern, batch_size=hparams.test_batch_size, mode=tf.estimator.ModeKeys.EVAL, hvd_info=None):
        return lambda: input_fn(
            input_pattern=pattern, metadata_path=hparams.metadata_path, batch_size=batch_size, mode=mode,
            vocab_table=vocab_utils.read_tf_vocab(hparams.vocab_file, hparams.UNK), hvd_info=hvd_info,
            vocab_table_for_id_ftr=vocab_utils.read_tf_vocab(hparams.vocab_file_for_id_ftr, hparams.UNK_FOR_ID_FTR),
            feature_names=hparams.feature_names, CLS=hparams.CLS, SEP=hparams.SEP, PAD=hparams.PAD,
            PAD_FOR_ID_FTR=hparams.PAD_FOR_ID_FTR, max_len=hparams.max_len, min_len=hparams.min_len,
            cnn_filter_window_size=max(hparams.filter_window_sizes) if hparams.ftr_ext == 'cnn' else 0)

    # Create TrainSpec for model training
    train_spec = tf.estimator.TrainSpec(
        input_fn=input_fn_common(hparams.train_file,
                                 batch_size=hparams.train_batch_size,
                                 mode=tf.estimator.ModeKeys.TRAIN,
                                 # Add horovod information if applicable
                                 hvd_info=hparams.hvd_info if hvd else None),
        hooks=bcast_hook,
        max_steps=hparams.num_train_steps)

    eval_spec = tf.estimator.EvalSpec(
        input_fn=input_fn_common(hparams.dev_file),
        exporters=exporter_list,
        steps=None,
        # Set throttle to 0 to start evaluation right away.
        # (Note: throttle_secs has to be 0 for horovod:
        # https://github.com/horovod/horovod/issues/182#issuecomment-533897757)
        throttle_secs=0,
        start_delay_secs=10)

    # Training and evaluation with dev set
    tf.estimator.train_and_evaluate(
        estimator=estimator,
        train_spec=train_spec,
        eval_spec=eval_spec)
    print("***** Training finished. *****")

    # Evaluation with test set: create an estimator with the best_checkpoint_dir to load the best model
    task_type = executor_utils.get_executor_task_type()
    do_evaluate = task_type == executor_utils.EVALUATOR or task_type == executor_utils.LOCAL_MODE
    if (not hvd and do_evaluate) or (hvd and hvd.rank() == 0):
        best_checkpoint_dir = path_join(hparams.out_dir, best_model_name)
        estimator_savedmodel = get_estimator(hparams, strategy=train_strategy, best_checkpoint=best_checkpoint_dir)
        result = estimator_savedmodel.evaluate(input_fn=input_fn_common(hparams.test_file))
        print("\n***** Evaluation on test set with best exported model: *****")
        for key in sorted(result.keys()):
            print("%s = %s" % (key, str(result[key])))


def get_estimator(hparams, strategy, best_checkpoint=None):
    config_kwargs = {
        'save_summary_steps': hparams.steps_per_stats,
        'save_checkpoints_steps': hparams.steps_per_eval,
        'log_step_count_steps': hparams.steps_per_stats,
        'keep_checkpoint_max': hparams.keep_checkpoint_max
    }
    # Handle sync distributed training case via use_horovod
    if hparams.use_horovod:
        import horovod.tensorflow as hvd
        # Adjust config for horovod
        session_config = tf.ConfigProto()
        session_config.allow_soft_placement = True
        # Pin each worker to a GPU
        session_config.gpu_options.visible_device_list = str(hvd.local_rank())
        config_kwargs['session_config'] = session_config

        model_dir = best_checkpoint if best_checkpoint is not None else hparams.out_dir if hvd.rank() == 0 else None
    # Handle single gpu or async distributed training case
    else:
        # TODO:
        # In the future we will support both sync training and async training on TOnY.
        # Now the sync training can run on kubernetes cluster using hvd lib.
        # 1. async training.
        # train_distribute=tf.contrib.distribute.ParameterServerStrategy()  #async strategy.
        # 2. sync strategy
        # train_distribute=tf.distribute.experimental.MultiWorkerMirroredStrategy() #sync strategy,
        # and remember to remove ps in tony params when use sync strategy,
        # Also note that MultiWorkerMirroredStrategy doesn't support bert-adam optimizer, but can use adam optimizer
        config_kwargs['train_distribute'] = strategy
        model_dir = best_checkpoint if best_checkpoint is not None else hparams.out_dir

    config = tf.estimator.RunConfig(**config_kwargs)
    # Build the estimator
    return tf.estimator.Estimator(model_fn=model_fn, model_dir=model_dir, params=hparams, config=config)


def serving_input_fn(hparams):
    """
    Creates a serving input function used by inference in model export.
    :param hparams: hparams
    :return: A ServingInputReceiver function that defines the inference requests and prepares for the model.
    """

    # Define placeholders and features
    doc_feature_names = [df for df in hparams.feature_names if df.startswith('doc_')]
    usr_feature_names = [df for df in hparams.feature_names if df.startswith('usr_')]
    doc_fields, doc_placeholders = train_helper.get_doc_fields(hparams)
    usr_fields, usr_placeholders = train_helper.get_usr_fields(hparams)

    doc_id_feature_names = [df for df in hparams.feature_names if df.startswith('docId_')]
    usr_id_feature_names = [df for df in hparams.feature_names if df.startswith('usrId_')]
    doc_id_fields, doc_id_placeholders = train_helper.get_doc_id_fields(hparams)
    usr_id_fields, usr_id_placeholders = train_helper.get_usr_id_fields(hparams)

    query, query_placeholder = train_helper.get_query(hparams)
    wide_ftr_placeholder, wide_ftrs = train_helper.create_placeholder_for_ftrs(
        "wide_ftr_placeholder", [None, hparams.num_wide], tf.float32, 'wide_ftrs', hparams.feature_names)
    wide_ftr_sp_idx_placeholder, wide_ftrs_sp_idx = train_helper.create_placeholder_for_ftrs(
        "wide_ftr_sp_idx_placeholder", [None, None], tf.int32, 'wide_ftrs_sp_idx', hparams.feature_names)
    wide_ftr_sp_val_placeholder, wide_ftrs_sp_val = train_helper.create_placeholder_for_ftrs(
        "wide_ftr_sp_val_placeholder", [None, None], tf.float32, 'wide_ftrs_sp_val', hparams.feature_names)
    uid_placeholder, uid = train_helper.create_placeholder_for_ftrs(
        "uid_placeholder", [], tf.int64, 'uid', hparams.feature_names, tf.constant([-1], dtype=tf.int64))
    weight_placeholder, weight = train_helper.create_placeholder_for_ftrs(
        "weight_placeholder", [], tf.float32, 'weight', hparams.feature_names, tf.constant([1.0], dtype=tf.float32))
    label_placeholder, label = train_helper.create_placeholder_for_ftrs(
        "label_placeholder", [None], tf.float32, 'label', hparams.feature_names)
    task_id_placeholder, task_id = train_helper.create_placeholder_for_ftrs(
        "task_id_placeholder", [], tf.int64, 'task_id', hparams.feature_names)
    # Placeholder tensors
    # Default uid as feature for detext integration, will be -1 by default if not present in data
    feature_placeholders = {'uid': uid_placeholder, 'weight': weight_placeholder, 'label': label_placeholder}
    # Features to feed into model (creating model_fn)
    features = {'uid': uid, 'weight': weight, 'label': label}
    # Add objects into pleaceholders and features
    for fname in hparams.feature_names:
        if fname == 'query':
            feature_placeholders[fname] = query_placeholder
            features[fname] = query
        elif fname.startswith('doc_'):
            if fname not in features:
                for fi in range(hparams.num_doc_fields):
                    features[doc_feature_names[fi]] = doc_fields[fi]
                    feature_placeholders[doc_feature_names[fi]] = doc_placeholders[fi]
        elif fname.startswith('usr_'):
            if fname not in features:
                for fi in range(hparams.num_usr_fields):
                    features[usr_feature_names[fi]] = usr_fields[fi]
                    feature_placeholders[usr_feature_names[fi]] = usr_placeholders[fi]
        elif fname.startswith('docId_'):
            if fname not in features:
                for fi in range(hparams.num_doc_id_fields):
                    features[doc_id_feature_names[fi]] = doc_id_fields[fi]
                    feature_placeholders[doc_id_feature_names[fi]] = doc_id_placeholders[fi]
        elif fname.startswith('usrId_'):
            if fname not in features:
                for fi in range(hparams.num_usr_id_fields):
                    features[usr_id_feature_names[fi]] = usr_id_fields[fi]
                    feature_placeholders[usr_id_feature_names[fi]] = usr_id_placeholders[fi]
        elif fname == 'wide_ftrs':
            feature_placeholders[fname] = wide_ftr_placeholder
            features[fname] = wide_ftrs
        elif fname == 'wide_ftrs_sp_val':
            feature_placeholders[fname] = wide_ftr_sp_val_placeholder
            features[fname] = wide_ftrs_sp_val
        elif fname == 'wide_ftrs_sp_idx':
            feature_placeholders[fname] = wide_ftr_sp_idx_placeholder
            features[fname] = wide_ftrs_sp_idx
        elif fname == 'task_id':
            feature_placeholders[fname] = task_id_placeholder
            features[fname] = task_id
        elif fname == 'label' or fname == 'weight' or fname == 'uid':
            continue
        else:
            raise ValueError("Unsupported feature_to_add argument: %s" % fname)

    return tf.estimator.export.ServingInputReceiver(
        features, feature_placeholders)


def model_fn(features, labels, mode, params):
    """
    Defines the model_fn to feed in to estimator
    :param features: dict containing the features in data
    :param labels: dict containing labels in data
    :param mode: running mode, in TRAIN/EVAL/PREDICT
    :param params: hparams used
    :return: tf.estimator.EstimatorSpec
    """
    query_field = features.get('query', None)

    uid = features.get('uid', None)

    weight = features.get('weight', None)
    wide_ftrs = features.get('wide_ftrs', None)

    wide_ftrs_sp_idx = features.get('wide_ftrs_sp_idx', None)
    wide_ftrs_sp_val = features.get('wide_ftrs_sp_val', None)

    doc_fields = [features[ftr_name] for ftr_name in features if ftr_name.startswith('doc_')]
    if len(doc_fields) == 0:
        doc_fields = None

    usr_fields = [features[ftr_name] for ftr_name in features if ftr_name.startswith('usr_')]
    if len(usr_fields) == 0:
        usr_fields = None

    doc_id_fields = [features[ftr_name] for ftr_name in features if ftr_name.startswith('docId_')]
    if len(doc_id_fields) == 0:
        doc_id_fields = None

    usr_id_fields = [features[ftr_name] for ftr_name in features if ftr_name.startswith('usrId_')]
    if len(usr_id_fields) == 0:
        usr_id_fields = None

    label_field = labels['label'] if mode != tf.estimator.ModeKeys.PREDICT else None
    labels_passthrough = features['label']

    group_size_field = features['group_size'] if mode != tf.estimator.ModeKeys.PREDICT else None

    # For multitask training
    task_id_field = features.get('task_id', None)  # shape=[batch_size,]

    # Update the weight with each task's weight such that weight per document = weight * task_weight
    if params.task_ids is not None:
        task_ids = params.task_ids  # e.g. [0, 1, 2]
        task_weights = params.task_weights  # e.g. [0.1, 0.3, 0.6]
        # Expand task_id_field with shape [batch_size, num_tasks]
        expanded_task_id_field = tf.transpose(tf.broadcast_to(task_id_field, [len(task_ids), tf.shape(task_id_field)[0]]))
        task_mask = tf.cast(tf.equal(expanded_task_id_field, task_ids), dtype=tf.float32)
        weight *= tf.reduce_sum(task_mask * task_weights, 1)  # shape=[batch_size,]

    # build graph
    model = DeepMatch(query=query_field,
                      wide_ftrs=wide_ftrs,
                      doc_fields=doc_fields,
                      usr_fields=usr_fields,
                      doc_id_fields=doc_id_fields,
                      usr_id_fields=usr_id_fields,
                      hparams=params,
                      mode=mode,
                      wide_ftrs_sp_idx=wide_ftrs_sp_idx,
                      wide_ftrs_sp_val=wide_ftrs_sp_val,
                      task_id_field=task_id_field)

    if mode == tf.estimator.ModeKeys.TRAIN:
        loss = compute_loss(params, model.scores, label_field, group_size_field, weight)
        train_op, _, _ = optimization.create_optimizer(params, loss)
        global_step = tf.train.get_global_step()
        train_tensors_log = {'loss': loss, 'global_step': global_step}
        logging_hook = tf.train.LoggingTensorHook(train_tensors_log, every_n_iter=10)
        return tf.estimator.EstimatorSpec(mode=mode,
                                          loss=loss,
                                          train_op=train_op,
                                          training_hooks=[logging_hook])

    elif mode == tf.estimator.ModeKeys.EVAL:
        loss = compute_loss(params, model.scores, label_field, group_size_field, weight)
        eval_metric_ops = {}
        for metric_name in params.all_metrics:
            metric_op_name = 'metric/{}'.format(metric_name)
            topk = int(metric_name.split('@')[1]) if '@' in metric_name else 10  # Default topk
            if metric_name.startswith('ndcg'):
                metric = metrics.compute_ndcg_tfr(model.scores, label_field, features, topk)
            elif metric_name.startswith('mrr'):
                metric = metrics.compute_mrr_tfr(model.scores, label_field, features)
            elif metric_name.startswith('precision'):
                metric = metrics.compute_precision_tfr(model.scores, label_field, features, topk)
            elif metric_name.startswith('traditional_ndcg'):
                metric = metrics.compute_ndcg(model.scores, label_field, group_size_field, topk)
            elif metric_name.startswith('li_mrr'):
                metric = metrics.compute_mrr(model.scores, labels['label'], features['group_size'], topk)
            elif metric_name == 'auc':
                metric = metrics.compute_auc(model.scores, label_field)
            elif metric_name == 'accuracy':
                metric = metrics.compute_accuracy(model.scores, label_field)
            elif metric_name == 'confusion_matrix':
                metric = metrics.compute_confusion_matrix(model.scores, label_field, params.num_classes)
            else:
                raise ValueError(f"Unsupported metrics: {metric_name}")
            eval_metric_ops[metric_op_name] = metric
        return tf.estimator.EstimatorSpec(mode,
                                          loss=loss,
                                          eval_metric_ops=eval_metric_ops)

    elif mode == tf.estimator.ModeKeys.PREDICT:
        # Prediction field for scoring models
        predictions = {
            'uid': uid,
            'scores': model.original_scores,
            'weight': weight,
            'label': labels_passthrough
        }
        # multiclass classification: export the probabilities across classes by applying softmax
        if params.num_classes > 1:
            predictions['multiclass_probabilities'] = tf.nn.softmax(model.scores)

        export_outputs = {
            'prediction': tf.estimator.export.PredictOutput(predictions)
        }
        # Provide an estimator spec for `ModeKeys.PREDICT` mode.
        return tf.estimator.EstimatorSpec(mode,
                                          predictions=predictions,
                                          export_outputs=export_outputs)
    else:
        raise ValueError("Only support mode as TRAIN/EVAL/PREDICT")


def compute_loss(hparams, scores, labels, group_size, weight):
    """ Computes ranking/classification loss with regularization """
    if weight is None:
        raise ValueError("weight should not be None")
    return compute_rank_clf_loss(hparams, scores, labels, group_size, weight) + tf.reduce_mean(
        weight) * compute_regularization_penalty(hparams)


def compute_rank_clf_loss(hparams, scores, labels, group_size, weight):
    """
    Compute ranking/classification loss
    Note that the tfr loss is slightly different than our implementation: the tfr loss is sum over all loss and
    devided by number of queries; our implementation is sum over all loss and devided by the number of larger than
    0 labels.
    """
    # Classification loss
    if hparams.num_classes > 1:
        labels = tf.cast(labels, tf.int32)
        labels = tf.squeeze(labels, -1)  # Last dimension is max_group_size, which should be 1
        return tf.losses.sparse_softmax_cross_entropy(logits=scores, labels=labels, weights=weight)

    # Expand weight to [batch size, 1] so that in inhouse ranking loss it can be multiplied with loss which is
    #   [batch_size, max_group_size]
    expanded_weight = tf.expand_dims(weight, axis=-1)

    # Ranking losses
    # tf-ranking loss
    if hparams.use_tfr_loss:
        weight_name = "weight"
        loss_fn = tfr.losses.make_loss_fn(hparams.tfr_loss_fn, lambda_weight=hparams.tfr_lambda_weights,
                                          weights_feature_name=weight_name)
        loss = loss_fn(labels, scores, {weight_name: expanded_weight})
        return loss

    # our own implementation
    if hparams.ltr_loss_fn == 'pairwise':
        lambdarank = LambdaRank()
        pairwise_loss, pairwise_mask = lambdarank(scores, labels, group_size)
        loss = tf.reduce_sum(tf.reduce_sum(pairwise_loss, axis=[1, 2]) * expanded_weight) / tf.reduce_sum(pairwise_mask)
    elif hparams.ltr_loss_fn == 'softmax':
        loss = compute_softmax_loss(scores, labels, group_size) * expanded_weight
        is_positive_label = tf.cast(tf.greater(labels, 0), dtype=tf.float32)
        loss = tf.div_no_nan(tf.reduce_sum(loss), tf.reduce_sum(is_positive_label))
    elif hparams.ltr_loss_fn == 'pointwise':
        loss = compute_sigmoid_cross_entropy_loss(scores, labels, group_size) * expanded_weight
        loss = tf.reduce_mean(loss)
    else:
        raise ValueError('Currently only support pointwise/pairwise/softmax/softmax_cls.')
    return loss



"""Utility functions for building models."""

import tensorflow as tf

from detext.train import data_fn
from detext.utils import vocab_utils


def create_placeholder_for_ftrs(ph_name, shape, dtype, ftr_name, ftr_names, default=None):
    """Creates placeholder and specifies batch_size=1"""
    ph = tf.placeholder(shape=shape, dtype=dtype, name=ph_name)
    ph_one_batch = tf.expand_dims(ph, axis=0) if ftr_name in ftr_names else default
    return ph, ph_one_batch


def get_query(hparams):
    """
    Helper function to get query and query_placeholder
    :param hparams: hparams
    :return: query and query_placeholder
    """
    # query text feature
    # If hparams.add_first_dim_for_query_placeholder is True, the query placeholder has dimension [None]
    # This is to use the query feature as a document feature in model serving
    if hparams.add_first_dim_for_query_placeholder:
        query_placeholder, query = create_placeholder_for_ftrs("query_placeholder", [None], tf.string, 'query',
                                                               hparams.feature_names)
    else:
        query_placeholder, query = create_placeholder_for_ftrs("query_placeholder", [], tf.string, 'query',
                                                               hparams.feature_names)
    if query is not None:
        if hparams.add_first_dim_for_query_placeholder:
            # remove added dimension
            query = tf.squeeze(query, [0])

        # tokenize query
        if hparams.regex_replace_pattern is not None:
            query = tf.regex_replace(input=query, pattern=hparams.regex_replace_pattern, rewrite=" \\1 ")

        query = data_fn.process_text(
            query,
            vocab_utils.read_tf_vocab(hparams.vocab_file, hparams.UNK),
            hparams.CLS, hparams.SEP, hparams.PAD,
            hparams.max_len,
            hparams.min_len,
            cnn_filter_window_size=max(hparams.filter_window_sizes) if hparams.ftr_ext == 'cnn' else 0
        )
    return query, query_placeholder


def get_doc_fields(hparams):
    """
    Each document field has a placeholder.
    The regex is to add whitespace on both sides of punctuations.
    :param hparams: hparams
    :param regex_replace_pattern: The regex pattern to add a white space before and after
    :return:
    """
    doc_text_placeholders = []
    doc_fields = []
    tf_vocab_table = vocab_utils.read_tf_vocab(hparams.vocab_file, hparams.UNK)
    for ftr_name in hparams.feature_names:
        if ftr_name.startswith('doc_'):
            # each document field is a placeholder (a string vector)
            placeholder = tf.placeholder(shape=[None], dtype=tf.string, name=ftr_name + "_placeholder")
            doc_text_placeholders.append(placeholder)

            one_doc_field = placeholder
            # add whitespace on both sides of punctuations if regex pattern is not None
            if hparams.regex_replace_pattern is not None:
                one_doc_field = tf.regex_replace(input=one_doc_field, pattern=hparams.regex_replace_pattern, rewrite=" \\1 ")
            one_doc_field = data_fn.process_text(
                one_doc_field,
                tf_vocab_table,
                hparams.CLS, hparams.SEP, hparams.PAD,
                hparams.max_len,
                hparams.min_len,
                cnn_filter_window_size=max(hparams.filter_window_sizes) if hparams.ftr_ext == 'cnn' else 0
            )
            one_doc_field = tf.expand_dims(one_doc_field, axis=0)
            doc_fields.append(one_doc_field)
    return doc_fields, doc_text_placeholders


def get_usr_fields(hparams):
    """
    Each user field has a placeholder.
    The regex is to add whitespace on both sides of punctuations.
    :param hparams: hparams
    :return:
    """
    usr_text_placeholders = []
    usr_fields = []
    tf_vocab_table = vocab_utils.read_tf_vocab(hparams.vocab_file, hparams.UNK)
    for ftr_name in hparams.feature_names:
        if ftr_name.startswith('usr_'):
            # If hparams.add_first_dim_for_usr_placeholder is True, the usr placeholders have dimension [None]
            # This is to use the usr field features as document features in model serving
            if hparams.add_first_dim_for_usr_placeholder:
                # each user field is a placeholder (one string)
                placeholder = tf.placeholder(shape=[None], dtype=tf.string, name=ftr_name + "_placeholder")
            else:
                placeholder = tf.placeholder(shape=[], dtype=tf.string, name=ftr_name + "_placeholder")
            usr_text_placeholders.append(placeholder)

            one_usr_field = placeholder
            # add whitespace on both sides of punctuations if regex pattern is not None
            if hparams.regex_replace_pattern is not None:
                one_usr_field = tf.regex_replace(input=one_usr_field, pattern=hparams.regex_replace_pattern, rewrite=" \\1 ")

            # remove added dimension
            if hparams.add_first_dim_for_usr_placeholder:
                one_usr_field = tf.squeeze(one_usr_field, [0])
            one_usr_field = tf.expand_dims(one_usr_field, axis=0)
            one_usr_field = data_fn.process_text(
                one_usr_field,
                tf_vocab_table,
                hparams.CLS, hparams.SEP, hparams.PAD,
                hparams.max_len,
                hparams.min_len,
                cnn_filter_window_size=max(hparams.filter_window_sizes) if hparams.ftr_ext == 'cnn' else 0
            )
            usr_fields.append(one_usr_field)
    return usr_fields, usr_text_placeholders


def get_doc_id_fields(hparams):
    """ Returns a list of processed doc id fields and the corresponding list of raw doc id placeholders
    Each document id field has a placeholder
    """
    doc_id_placeholders = []
    doc_fields = []
    tf_vocab_table = vocab_utils.read_tf_vocab(hparams.vocab_file_for_id_ftr, hparams.UNK_FOR_ID_FTR)
    for ftr_name in hparams.feature_names:
        if ftr_name.startswith('docId_'):
            # each document id field is a placeholder (a string vector)
            placeholder = tf.placeholder(shape=[None], dtype=tf.string, name=ftr_name + "_placeholder")
            doc_id_placeholders.append(placeholder)

            one_doc_field = placeholder
            one_doc_field = data_fn.process_id(
                one_doc_field,
                tf_vocab_table,
                hparams.PAD_FOR_ID_FTR
            )
            one_doc_field = tf.expand_dims(one_doc_field, axis=0)
            doc_fields.append(one_doc_field)
    return doc_fields, doc_id_placeholders


def get_usr_id_fields(hparams):
    """ Returns a list of processed usr id fields and the corresponding list of raw usr id placeholders
    Each user field has a placeholder
    """
    usr_id_placeholders = []
    usr_fields = []
    tf_vocab_table = vocab_utils.read_tf_vocab(hparams.vocab_file_for_id_ftr, hparams.UNK_FOR_ID_FTR)
    for ftr_name in hparams.feature_names:
        if ftr_name.startswith('usrId_'):
            # each user id field is a placeholder (one string)
            placeholder = tf.placeholder(shape=[], dtype=tf.string, name=ftr_name + "_placeholder")
            usr_id_placeholders.append(placeholder)

            one_usr_field = placeholder

            one_usr_field = tf.expand_dims(one_usr_field, axis=0)
            one_usr_field = data_fn.process_id(
                one_usr_field,
                tf_vocab_table,
                hparams.PAD_FOR_ID_FTR
            )

            usr_fields.append(one_usr_field)
    return usr_fields, usr_id_placeholders



"""
Representation-based models.
"""
import tensorflow as tf
from functools import partial

from detext.model import bert_model
from detext.model import cnn_model
from detext.model import id_embed_model
from detext.model import lstm_lm_model
from detext.model import lstm_model


class RepModel(object):
    """
    The representation-based model to generate deep features, based on the textual and id inputs: query, user fields,
    document fields.

    Attributes:
        doc_ftrs  Tensor Shape=[batch_size, max_group_size, num_doc_fields, ftr_size]
        usr_ftrs  Tensor Shape=[batch_size, num_usr_fields, ftr_size]
        num_doc_fields int Number of document fields (text + id)
        num_usr_fields int Number of document fields (text + id)
        num_sim_ftrs  int Number of similarity features
        sim_ftrs  Tensor Shape=[batch_size, max_group_size, num_sim_ftrs] Similarity features
    """

    def __init__(self,
                 query,
                 doc_fields,
                 usr_fields,
                 doc_id_fields,
                 usr_id_fields,
                 hparams,
                 mode):
        """
        :param query: Tensor  Input query. Shape=[batch_size, query_length]
        :param doc_fields: list(Tensor) or a pre-computed Tensor  List of document fields. Each has
            shape=[batch_size, max_group_size, doc_field_length]
        :param usr_fields: list(Tensor) or a pre-computed Tensor  List of user fields. Each has
            shape=[batch_size, usr_field_length]
        :param hparams: HParams  Hyper parameters
        :param mode: TRAIN/EVAL/INFER
        """
        self._query = query
        self._hparams = hparams
        self._mode = mode
        self._usr_fields = usr_fields
        self._doc_fields = doc_fields
        self._usr_id_fields = usr_id_fields
        self._doc_id_fields = doc_id_fields

        # Apply text encoder on query/doc/user fields
        ftr_ext_name2model = {
            'cnn': cnn_model.CnnModel,
            'bert': bert_model.BertModel,
            'lstm_lm': lstm_lm_model.LstmLmModel,
            'lstm': lstm_model.LstmModel
        }
        self.text_encoding_model = ftr_ext_name2model[hparams.ftr_ext](query, doc_fields, usr_fields, hparams, mode)

        # Apply embedding lookup and averaging for usr/doc id features
        self.id_encoding_model = id_embed_model.IdEmbedModel(
            doc_id_fields=self._doc_id_fields,
            usr_id_fields=self._usr_id_fields,
            hparams=self._hparams,
            mode=self._mode
        )

        # Use text_ftr_size as the size of representation for id embedding and text embedding
        self.ftr_size = self.text_encoding_model.text_ftr_size

        # Add text ftrs and id ftrs together
        for field in ['query_ftrs', 'usr_ftrs', 'doc_ftrs']:
            text_field_tensor = getattr(self.text_encoding_model, field, None)
            id_field_tensor = getattr(self.id_encoding_model, field, None)

            # Project text embedding to representation space
            # if text_field_tensor is not None:
            #     text_field_tensor = tf.layers.dense(text_field_tensor, units=self.ftr_size, use_bias=True,
            #                                         name="text_projection", reuse=tf.AUTO_REUSE)

            # Project id embedding to representation space
            if id_field_tensor is not None:
                id_field_tensor = tf.layers.dense(id_field_tensor, units=self.ftr_size, use_bias=True,
                                                  name="id_projection", reuse=tf.AUTO_REUSE)

            # Compose a new tensor to include user id info and user text info
            # I.e., if there are 2 user id fields, 3 user text fields, we have 2 + 3 = 5 user embedding fields
            # Same processing also works for doc fields
            if text_field_tensor is not None and id_field_tensor is not None:
                composite_field_tensor = tf.concat([text_field_tensor, id_field_tensor], axis=-2)
            elif text_field_tensor is not None:
                composite_field_tensor = text_field_tensor
            else:
                composite_field_tensor = id_field_tensor

            # Assign a name to each text feature in the graph so that we can quickly locate them in the graph
            if composite_field_tensor is not None:
                composite_field_tensor = tf.identity(composite_field_tensor, name=field)
            setattr(self, field, composite_field_tensor)

        self.num_doc_fields = hparams.num_doc_fields + hparams.num_doc_id_fields
        self.num_usr_fields = hparams.num_usr_fields + hparams.num_usr_id_fields

        self.num_sim_ftrs, self.sim_ftrs = self.compute_sim_ftrs()

    def get_data_size(self):
        """Infers batch size, max group size"""
        if self.doc_ftrs is not None:
            data = self.doc_ftrs
        elif self.query_ftrs:
            data = self.query_ftrs
        elif self.usr_ftrs:
            data = self.usr_ftrs
        else:
            raise ValueError('Cannot infer data size.')
        data_shape = tf.shape(data)
        return data_shape[0], data_shape[1]

    def compute_sim_ftrs(self):
        """
        Computes the interaction features between user/document features and query features.
        When no query and user fields, then return doc_ftrs directly (as concatenation)

        Supported interaction modes (hparams.emb_sim_func) are:
            1. inner: computes cosine similarity between user fields/query and each document field
                num_sim_ftrs=num_doc_fields
            2. concat: concatenates user fields/query and each document field
                num_sim_ftrs=self.ftr_size * (num_usr_fields + num_doc_fields)
            3. hadamard: computes Hadamard product between user fields/query and each document field
                num_sim_ftrs=self.ftr_size * num_doc_fields * num_usr_fields

        :return num_sim_ftrs,sim_ftrs  Shape of sim_ftrs is [batch_size, group_size, num_sim_ftrs]. The value of
            num_sim_ftrs is as shown in the instructions above
        """
        hparams = self._hparams

        batch_size, max_group_size = self.get_data_size()
        num_doc_fields = self.num_doc_fields
        ftr_size = self.ftr_size

        query_ftrs = self.query_ftrs  # [batch_size, ftr_size]
        usr_ftrs = self.usr_ftrs  # [batch_size, num_usr_fields, ftr_size]
        doc_ftrs = self.doc_ftrs  # [batch_size, group_size, num_doc_fields, ftr_size]

        # If query is None, use doc_ftrs only and no interaction is performed
        if query_ftrs is None and usr_ftrs is None:
            sim_ftrs = tf.reshape(doc_ftrs, shape=[batch_size, max_group_size, ftr_size * num_doc_fields])
            num_sim_ftrs = ftr_size * num_doc_fields
            return num_sim_ftrs, sim_ftrs

        # If use_doc_projection, then the n doc fields are projected to 1 vector space
        if hparams.use_doc_projection:
            doc_ftrs = tf.reshape(doc_ftrs, shape=[batch_size, max_group_size, 1, ftr_size * num_doc_fields])
            doc_ftrs = tf.layers.dense(doc_ftrs,
                                       ftr_size,
                                       use_bias=True,
                                       activation=tf.tanh,
                                       name="doc_ftrs_projection_layer")  # [batch_size, max_group_size, 1, ftr_size]
            doc_ftrs = tf.identity(doc_ftrs, name='doc_ftrs_projection')
            num_doc_fields = 1

        # Compute interaction between user text/query and document text
        num_usr_fields = 0
        tot_usr_ftrs = []
        if usr_ftrs is not None:
            # If use_usr_projection, then the n usr fields are projected to 1 vector space
            if hparams.use_usr_projection:
                usr_ftrs = tf.reshape(usr_ftrs, shape=[batch_size, 1, ftr_size * self.num_usr_fields])
                usr_ftrs = tf.layers.dense(usr_ftrs,
                                           ftr_size,
                                           use_bias=True,
                                           activation=tf.tanh,
                                           name="usr_ftrs_projection_layer")  # [batch_size, 1, ftr_size]
                usr_ftrs = tf.identity(usr_ftrs, name='usr_ftrs_projection')
                num_usr_fields = 1
            else:
                num_usr_fields += self.num_usr_fields
            tot_usr_ftrs.append(usr_ftrs)

        # Treat query as one user field and append it to user field
        if query_ftrs is not None:
            query_ftrs = tf.expand_dims(query_ftrs, axis=1)  # [batch_size, 1, ftr_size]
            tot_usr_ftrs.append(query_ftrs)
            num_usr_fields += 1
        usr_ftrs = tf.concat(tot_usr_ftrs, axis=1)  # [batch_size, num_usr_fields, ftr_size]

        sim_ftrs, num_sim_ftrs = self.compute_sim_ftrs_for_usr_doc(doc_ftrs, usr_ftrs, num_doc_fields,
                                                                   num_usr_fields,
                                                                   hparams.emb_sim_func)
        return num_sim_ftrs, sim_ftrs

    def compute_sim_ftrs_for_usr_doc(self, doc_ftrs, usr_ftrs, num_doc_fields, num_usr_fields, sim_func_names):
        """Computes similarity between usr_ftrs and doc_ftrs with all function names specified in sim_func_names and
            concatenates the results

        :param doc_ftrs Shape=[batch_size, group_size, num_doc_fields, ftr_size]
        :param usr_ftrs Shape=[batch_size, num_usr_fields, ftr_size]
        :param sim_func_names Similarity function names
        """
        sim_ftrs_list = list()
        num_sim_ftrs_list = list()

        name2func = {'inner': compute_inner_ftrs_for_usr_doc,
                     'concat': compute_concat_ftrs_for_usr_doc,
                     'hadamard': compute_hadamard_prod_for_usr_doc,
                     'diff': compute_elem_diff_for_usr_doc}

        for name in sim_func_names:
            sim_ftrs, num_sim_ftrs = name2func[name](doc_ftrs, usr_ftrs, num_doc_fields, num_usr_fields,
                                                     self.ftr_size)
            sim_ftrs_list.append(sim_ftrs)
            num_sim_ftrs_list.append(num_sim_ftrs)

        return tf.concat(sim_ftrs_list, axis=-1), sum(num_sim_ftrs_list)


def compute_elem_diff_for_usr_doc(doc_ftrs, usr_ftrs, num_doc_fields, num_usr_fields, num_deep):
    """Computes Elementwise difference between usr_field_ftrs and doc_ftrs

    :param doc_ftrs Document features. Shape=[batch_size, group_size, num_doc_fields, num_deep]
    :param usr_field_ftrs Shape=[batch_size, num_usr_fields, num_deep]
    :return (sim_ftrs, num_sim_ftrs). Similarity features and the number of similarity features (the last dimension
    of sim_ftrs)
    """

    def compute_diff(usr_field_ftrs, doc_ftrs):
        """ Computes Elementwise diff between usr_field_ftrs and doc_ftrs

        :param doc_ftrs Document features. Shape=[batch_size, group_size, num_doc_fields, num_deep]
        :param usr_field_ftrs Shape=[batch_size, num_deep]
        """
        usr_field_ftrs = tf.expand_dims(tf.expand_dims(usr_field_ftrs, axis=1), axis=1)  # [batch_size, 1, 1, num_deep]
        elementwise_diff = usr_field_ftrs - doc_ftrs  # [batch_size, group_size, num_doc_fields, num_deep]
        return elementwise_diff  # [batch_size, group_size, num_doc_fields, num_deep]

    batch_size = tf.shape(doc_ftrs)[0]
    max_group_size = tf.shape(doc_ftrs)[1]

    # Shape=[num_usr_fields, batch_size, group_size, num_doc_fields, num_deep]
    sim_ftrs = tf.map_fn(partial(compute_diff, doc_ftrs=doc_ftrs), tf.transpose(usr_ftrs, [1, 0, 2]))
    # Shape=[batch_size, group_size, num_doc_fields, num_usr_fields, num_deep]
    sim_ftrs = tf.transpose(sim_ftrs, [1, 2, 3, 4, 0])

    num_sim_ftrs = num_doc_fields * num_usr_fields * num_deep
    # Shape=[batch_size, group_size, num_sim_ftrs]
    sim_ftrs = tf.reshape(sim_ftrs, [batch_size, max_group_size, num_sim_ftrs])
    return sim_ftrs, num_sim_ftrs


def compute_hadamard_prod_for_usr_doc(doc_ftrs, usr_ftrs, num_doc_fields, num_usr_fields, num_deep):
    """Computes Hadamard product between usr_field_ftrs and doc_ftrs

    :param doc_ftrs Document features. Shape=[batch_size, group_size, num_doc_fields, num_deep]
    :param usr_field_ftrs Shape=[batch_size, num_usr_fields, num_deep]
    :return (sim_ftrs, num_sim_ftrs). Similarity features and the number of similarity features (the last dimension
    of sim_ftrs)
    """

    def compute_hadamard(usr_field_ftrs, doc_ftrs):
        """ Computes Hadamard product between usr_field_ftrs and doc_ftrs

        :param doc_ftrs Document features. Shape=[batch_size, group_size, num_doc_fields, num_deep]
        :param usr_field_ftrs Shape=[batch_size, num_deep]
        """
        usr_field_ftrs = tf.expand_dims(tf.expand_dims(usr_field_ftrs, axis=1), axis=1)  # [batch_size, 1, 1, num_deep]
        hadamard_prod = usr_field_ftrs * doc_ftrs  # [batch_size, group_size, num_doc_fields, num_deep]
        return hadamard_prod  # [batch_size, group_size, num_doc_fields, num_deep]

    batch_size = tf.shape(doc_ftrs)[0]
    max_group_size = tf.shape(doc_ftrs)[1]

    # Shape=[num_usr_fields, batch_size, group_size, num_doc_fields, num_deep]
    sim_ftrs = tf.map_fn(partial(compute_hadamard, doc_ftrs=doc_ftrs), tf.transpose(usr_ftrs, [1, 0, 2]))
    # Shape=[batch_size, group_size, num_doc_fields, num_usr_fields, num_deep]
    sim_ftrs = tf.transpose(sim_ftrs, [1, 2, 3, 4, 0])

    num_sim_ftrs = num_doc_fields * num_usr_fields * num_deep
    # Shape=[batch_size, group_size, num_sim_ftrs]
    sim_ftrs = tf.reshape(sim_ftrs, [batch_size, max_group_size, num_sim_ftrs])
    return sim_ftrs, num_sim_ftrs


def compute_concat_ftrs_for_usr_doc(doc_ftrs, usr_ftrs, num_doc_fields, num_usr_fields, num_deep):
    """Concatenates doc_ftrs and usr_ftrs

    :param doc_ftrs Shape=[batch_size, group_size, num_doc_fields, num_deep]
    :param usr_ftrs Shape=[batch_size, num_usr_fields, num_deep]
    :return (sim_ftrs, num_sim_ftrs). Similarity features and the number of similarity features (the last dimension
    of sim_ftrs)
    """
    batch_size = tf.shape(doc_ftrs)[0]
    max_group_size = tf.shape(doc_ftrs)[1]

    doc_ftrs = tf.reshape(doc_ftrs, shape=[batch_size, max_group_size, num_deep * num_doc_fields])

    usr_ftrs = tf.reshape(usr_ftrs, shape=[batch_size, 1, -1]) + tf.zeros([
        batch_size, max_group_size, num_deep * num_usr_fields])

    sim_ftrs = tf.concat([doc_ftrs, usr_ftrs], axis=2)
    num_sim_ftrs = num_deep * (num_doc_fields + num_usr_fields)
    return sim_ftrs, num_sim_ftrs


def compute_inner_ftrs_for_usr_doc(doc_ftrs, usr_ftrs, num_doc_fields, num_usr_fields, num_deep):
    """Computes cosine similarity between user and doc fields.

    :param doc_ftrs Shape=[batch_size, group_size, num_doc_fields, num_deep]
    :param usr_ftrs Shape=[batch_size, num_usr_fields, num_deep]
    :return (sim_ftrs, num_sim_ftrs). Similarity features and the number of similarity features, which is the last
        dimension of sim_ftrs
    """

    def compute_inner_sim(usr_field_ftrs, doc_ftrs):
        """Computes cosine similarity score between usr_field_ftrs and doc_ftrs

        :param doc_ftrs Document features. Shape=[batch_size, group_size, num_doc_fields, num_deep]
        :param usr_field_ftrs Shape=[batch_size, num_deep]
        """
        usr_field_ftrs = tf.expand_dims(tf.expand_dims(usr_field_ftrs, axis=1), axis=1)
        sim_ftrs = tf.reduce_sum(usr_field_ftrs * doc_ftrs, axis=-1)
        return sim_ftrs  # [batch_size, group_size, num_doc_fields]

    batch_size = tf.shape(doc_ftrs)[0]
    max_group_size = tf.shape(doc_ftrs)[1]

    doc_ftrs = tf.nn.l2_normalize(doc_ftrs, axis=-1)
    usr_ftrs = tf.nn.l2_normalize(usr_ftrs, axis=-1)
    # Shape=[num_usr_fields, batch_size, group_size, num_doc_fields]
    sim_ftrs = tf.map_fn(partial(compute_inner_sim, doc_ftrs=doc_ftrs), tf.transpose(usr_ftrs, [1, 0, 2]))
    # Shape=[batch_size, group_size, num_doc_fields, num_usr_fields]
    sim_ftrs = tf.transpose(sim_ftrs, [1, 2, 3, 0])
    num_sim_ftrs = num_doc_fields * num_usr_fields
    # Shape=[batch_size, group_size, num_doc_fields * num_usr_fields]
    sim_ftrs = tf.reshape(sim_ftrs, [batch_size, max_group_size, num_sim_ftrs])
    return sim_ftrs, num_sim_ftrs



"""
Data processing.  Two important functions:
1. training_input_fn(), used during training and evaluation.
2. serving_input_receiver_fn(), used by model serving.
"""

import tensorflow as tf
import tensorflow_ranking as tfr
from functools import partial

from detext.utils.misc_utils import get_input_files

_LABEL_PADDING = tfr.data._PADDING_LABEL


def _get_padded_shapes_and_values(feature_names, PAD, vocab_table, PAD_FOR_ID_FTR, vocab_table_for_id_ftr):
    """Returns padded_shape and padd_values for each feature in feature_names

    :param feature_names List of feature names EXCLUDING 'label'
    :param pad_id Id of padding token in the vocab table. Used as the padding value for text fields.
    """
    padded_shapes = {'group_size': tf.TensorShape([])}
    padded_values = {'group_size': 0}

    pad_id = tf.cast(vocab_table.lookup(tf.constant(PAD)), tf.int32)

    if vocab_table_for_id_ftr is not None:
        pad_id_for_id_ftr = tf.cast(vocab_table_for_id_ftr.lookup(tf.constant(PAD_FOR_ID_FTR)), tf.int32)

    for name in feature_names:
        if name == 'query' or name.startswith('usr_'):
            padded_shapes[name] = tf.TensorShape([None])
            padded_values[name] = pad_id
        elif name.startswith('doc_'):
            padded_shapes[name] = tf.TensorShape([None, None])
            padded_values[name] = pad_id
        elif name.startswith('usrId_'):
            assert vocab_table_for_id_ftr is not None, "Need to provide vocab_table_for_id_ftr"
            padded_shapes[name] = tf.TensorShape([None])
            padded_values[name] = pad_id_for_id_ftr
        elif name.startswith('docId_'):
            assert vocab_table_for_id_ftr is not None, "Need to provide vocab_table_for_id_ftr"
            padded_shapes[name] = tf.TensorShape([None, None])
            padded_values[name] = pad_id_for_id_ftr
        elif name == 'weight':
            padded_shapes[name] = tf.TensorShape([])
            padded_values[name] = 0.0
        elif name in ('uid', 'task_id'):
            padded_shapes[name] = tf.TensorShape([])
            padded_values[name] = tf.cast(0, tf.int64)
        elif name == 'wide_ftrs':
            padded_shapes[name] = tf.TensorShape([None, None])
            padded_values[name] = 0.0
        elif name == 'wide_ftrs_sp_val':
            padded_shapes[name] = tf.TensorShape([None, None])
            padded_values[name] = 1.0
        elif name == 'wide_ftrs_sp_idx':
            padded_shapes[name] = tf.TensorShape([None, None])
            padded_values[name] = 0
        elif name == 'label':
            padded_shapes[name] = tf.TensorShape([None])
            padded_values[name] = _LABEL_PADDING
        else:
            raise KeyError('Feature name ({}) not supported'.format(name))

    padded_shapes = (padded_shapes, {'label': tf.TensorShape([None])})
    padded_values = (padded_values, {'label': _LABEL_PADDING})  # use -1 as padding for labels
    return padded_shapes, padded_values


def _get_tfrecord_feature_parsing_schema(feature_names):
    """Returns parsing schema for input TFRecord

    :param feature_names: List of feature names INCLUDING 'label'
    """
    features_tfr = dict()
    for name in feature_names:
        if name == 'query' or name.startswith('usr_') or name.startswith('usrId_'):
            features_tfr[name] = tf.FixedLenFeature(shape=[1], dtype=tf.string)
        elif name == 'weight':
            features_tfr[name] = tf.FixedLenFeature(shape=[1], dtype=tf.float32)
            # Default uid as feature for detext integration, will be -1 by default if not present in data
        elif name in ('uid', 'task_id'):
            features_tfr[name] = tf.FixedLenFeature(shape=[1], dtype=tf.int64)
        elif name.startswith(('doc_', 'docId_')):
            features_tfr[name] = tf.VarLenFeature(dtype=tf.string)
        elif name in ('wide_ftrs', 'label', 'wide_ftrs_sp_val'):
            features_tfr[name] = tf.VarLenFeature(dtype=tf.float32)
        elif name in ('wide_ftrs_sp_idx',):
            features_tfr[name] = tf.VarLenFeature(dtype=tf.int64)
        else:
            raise KeyError("Unknown feature in tfrecord feature parsing: %s" % name)
    return features_tfr


def _cast_to_dtype_of_smaller_size(t):
    """Casts tensor to smaller storage dtype. int64 -> int32, float64 -> float32"""
    if t.dtype == tf.int64:
        return tf.to_int32(t)
    elif t.dtype == tf.float64:
        return tf.to_float(t)
    else:
        return t


def _convert_ftrs_to_dense_tensor(t, name):
    """Converts the VarLenFeature to dense format"""
    if name == 'wide_ftrs_sp_val':
        t = tf.sparse_tensor_to_dense(t, default_value=1.)  # For sparse features values, padding should be 1
    elif name.startswith('doc_') or name.startswith('docId_'):
        t = tf.sparse_tensor_to_dense(t, default_value='')
    elif name == 'wide_ftrs':
        t = tf.sparse_tensor_to_dense(t, default_value=0.)
    elif name == 'label':
        t = tf.sparse_tensor_to_dense(t, default_value=_LABEL_PADDING)
    elif name == 'wide_ftrs_sp_idx':  # If there's sparse wide features, it's required that 0 is only used for padding
        t = tf.sparse_tensor_to_dense(t, default_value=0)
    else:
        pass
    return t


def _reshape_ftrs_to_group_wise(features, feature_names, num_docs):
    """Reshapes features from [num_fts] to [num_ftrs_each_doc, num_docs]

    :param features Mapping from feature name to tensor
    """
    if 'wide_ftrs' in feature_names:
        features['wide_ftrs'] = tf.reshape(features['wide_ftrs'], shape=[num_docs, -1])
    if 'wide_ftrs_sp_idx' in feature_names:
        features['wide_ftrs_sp_idx'] = tf.reshape(features['wide_ftrs_sp_idx'], shape=[num_docs, -1])
    if 'wide_ftrs_sp_val' in feature_names:
        features['wide_ftrs_sp_val'] = tf.reshape(features['wide_ftrs_sp_val'], shape=[num_docs, -1])
    return features


def process_text(text,
                 vocab_table,
                 CLS, SEP, PAD,
                 max_len=None,
                 min_len=None,
                 cnn_filter_window_size=0):
    """
    Text processing: add cls/sep, handle max_len/min_len, etc.
    Text must be a string vector.
    """
    # add cls and sep
    num_docs = tf.shape(text)[0]
    cls_vec = tf.tile([CLS], [num_docs])
    sep_vec = tf.tile([SEP], [num_docs])
    if cnn_filter_window_size <= 2:
        text = tf.reduce_join([cls_vec, text, sep_vec], 0, separator=' ')
    else:
        # make sure each word appears the same times in CNN filters
        to_join = []
        for _ in range(cnn_filter_window_size - 1):
            to_join.append(cls_vec)
        to_join.append(text)
        for _ in range(cnn_filter_window_size - 1):
            to_join.append(sep_vec)
        text = tf.reduce_join(to_join, 0, separator=' ')

    # Split by spaces, and generate dense matrix
    text_sp = tf.string_split(text)
    text = tf.sparse_tensor_to_dense(text_sp, default_value=PAD)

    # handle max_len
    if max_len:
        text = text[:, :max_len]
    # handle min_len
    if min_len:
        text = tf.cond(
            tf.shape(text)[1] >= min_len,
            lambda: text,
            lambda: tf.pad(text, [[0, 0], [0, min_len - tf.shape(text)[1]]], constant_values=PAD))

    # Convert the word strings to ids.
    text = tf.cast(vocab_table.lookup(text), tf.int32)

    return text


def process_id(input_ids,
               vocab_table,
               PAD,
               max_len=None,
               min_len=None):
    """Converts input ids (in text string) into their indices in vocab

    :param input_ids Tensor Id features. Shape=[group_size]
    :param vocab_table TFLookupTable Vocab table for id features
    :param PAD Padding token for id features
    :param max_len Maximum number of id features for one doc
    :param min_len Minimum number of id features for one doc
    """
    # Split by spaces, and generate dense matrix
    input_ids_sp = tf.string_split(input_ids)
    input_ids = tf.sparse_tensor_to_dense(input_ids_sp, default_value=PAD)

    # handle max_len
    if max_len:
        input_ids = input_ids[:, :max_len]

    # handle min_len
    if min_len:
        input_ids = tf.cond(
            tf.shape(input_ids)[1] >= min_len,
            lambda: input_ids,
            lambda: tf.pad(input_ids, [[0, 0], [0, min_len - tf.shape(input_ids)[1]]], constant_values=PAD))

    # Convert input id strings to indices
    input_ids = tf.cast(vocab_table.lookup(input_ids), tf.int32)

    return input_ids


def input_fn(input_pattern, metadata_path, batch_size, mode,
             vocab_table, vocab_table_for_id_ftr,
             feature_names,
             CLS, SEP, PAD, PAD_FOR_ID_FTR,
             max_len=None,
             min_len=None,
             cnn_filter_window_size=0,
             prefetch_size=100,
             block_length=100,
             hvd_info=None):
    """ Returns a dataset from avro/tfrecord file """
    return input_fn_tfrecord(input_pattern, batch_size, mode,
                             vocab_table, vocab_table_for_id_ftr,
                             feature_names,
                             CLS, SEP, PAD, PAD_FOR_ID_FTR,
                             max_len, min_len,
                             cnn_filter_window_size,
                             block_length,
                             prefetch_size,
                             hvd_info=hvd_info)


def input_fn_tfrecord(input_pattern,
                      batch_size,
                      mode,
                      vocab_table, vocab_table_for_id_ftr,
                      feature_names,
                      CLS, SEP, PAD, PAD_FOR_ID_FTR,
                      max_len=None,
                      min_len=None,
                      cnn_filter_window_size=0,
                      block_length=100,
                      prefetch_size=100,
                      num_data_process_threads=32,
                      hvd_info=None):
    """
    Data input function for training given TFRecord
    """
    output_buffer_size = 1000

    input_files = get_input_files(input_pattern)
    feature_names = list(feature_names)
    if len(input_files) > 1:  # Multiple input files
        # Preprocess files concurrently, and interleave blocks of block_length records from each file
        dataset = tf.data.Dataset.from_tensor_slices(tf.constant(input_files))
        # shard input when using horovod
        if hvd_info is not None:
            dataset = dataset.shard(hvd_info['size'], hvd_info['rank'])
        dataset = dataset.shuffle(buffer_size=len(input_files))
        hvd_size = hvd_info['size'] if hvd_info is not None else 1
        # `cycle_length` is the number of parallel files that get read.
        cycle_length = min(num_data_process_threads, int(len(input_files) / hvd_size))
        dataset = dataset.interleave(tf.data.TFRecordDataset, cycle_length=cycle_length,
                                     block_length=block_length)
    else:
        dataset = tf.data.TFRecordDataset(input_files[0])

    dataset = tfrecord_transform_fn(dataset,
                                    batch_size,
                                    mode,
                                    vocab_table, vocab_table_for_id_ftr,
                                    feature_names,
                                    CLS, SEP, PAD, PAD_FOR_ID_FTR,
                                    output_buffer_size,
                                    max_len,
                                    min_len,
                                    cnn_filter_window_size,
                                    prefetch_size,
                                    num_data_process_threads)
    return dataset


def tfrecord_transform_fn(dataset,
                          batch_size,
                          mode,
                          vocab_table, vocab_table_for_id_ftr,
                          feature_names,
                          CLS, SEP, PAD, PAD_FOR_ID_FTR,
                          output_buffer_size,
                          max_len=None,
                          min_len=None,
                          cnn_filter_window_size=0,
                          prefetch_size=100,
                          num_data_process_threads=32):
    if mode == tf.estimator.ModeKeys.TRAIN:
        dataset = dataset.shuffle(output_buffer_size)
        dataset = dataset.repeat()

    def process_data(record, features_schema):
        # parse tfrecord
        example = tf.parse_single_example(record, features_schema)

        # process each field
        features = dict()
        for name in feature_names:
            t = example[name]
            if name not in ('uid', 'task_id'):
                t = _cast_to_dtype_of_smaller_size(t)
            t = _convert_ftrs_to_dense_tensor(t, name)

            # for text data, call process_text()
            if name == 'query' or name.startswith('doc_') or name.startswith('usr_'):
                t = process_text(t, vocab_table, CLS, SEP, PAD, max_len, min_len, cnn_filter_window_size)

            if name.startswith('usrId_') or name.startswith('docId_'):
                t = process_id(t, vocab_table_for_id_ftr, PAD_FOR_ID_FTR)

            if name in ('query', 'weight', 'uid', 'task_id') or name.startswith(('usr_', 'usrId_')):
                t = tf.squeeze(t, axis=0)

            features[name] = t

        num_docs = tf.size(features['label'])
        # get group size from label
        features['group_size'] = num_docs
        features = _reshape_ftrs_to_group_wise(features, feature_names, num_docs)
        features.setdefault('weight', tf.constant(1.0, dtype=tf.float32))
        # Default uid as feature for detext integration, will be -1 by default if not present in data
        features.setdefault('uid', tf.constant(-1, dtype=tf.int64))

        label = {'label': tf.to_float(features['label'])}
        return features, label

    features_schema = _get_tfrecord_feature_parsing_schema(feature_names)
    # process data
    dataset = dataset.map(partial(process_data, features_schema=features_schema),
                          num_parallel_calls=num_data_process_threads)

    if 'weight' not in feature_names:
        feature_names.append('weight')
    if 'uid' not in feature_names:
        feature_names.append('uid')
    padded_shapes, padded_values = _get_padded_shapes_and_values(feature_names, PAD, vocab_table,
                                                                 PAD_FOR_ID_FTR, vocab_table_for_id_ftr)

    dataset = (dataset
               .padded_batch(batch_size, padded_shapes=padded_shapes, padding_values=padded_values)
               .prefetch(prefetch_size))
    return dataset



"""
The model that uses CNN to extract features from texts, and generate a score for a query/doc pair
"""

import tensorflow as tf

from detext.utils import model_utils


class CnnModel(object):

    def __init__(self,
                 query,
                 doc_fields,
                 usr_fields,
                 hparams,
                 mode):
        """
        Applies CNN to convert text to a fix length embedding

        :param query: Tensor(dtype=tf.int)  Shape=[batch_size, query_length]
        :param doc_fields: list(Tensor(dtype=int))/Tensor  A list of document fields. Each has shape=
            [batch_size, max_group_size, doc_field_length]. For online scoring, these fields may be precomputed and
            input as Tensor
        :param usr_fields: list(Tensor(dtype=int))/Tensor  A list of user fields. Each has shape=
            [batch_size, usr_field_length]. For online scoring, these fields may be precomputed and
            input as Tensor
        :param hparams: HParams  Important CNN params are: filter_window_sizes:list(int), num_filters:int
        :param mode: TRAIN/EVAL/INFER
        """
        self._query = query
        self._doc_fields = doc_fields
        self._usr_fields = usr_fields
        self._hparams = hparams
        self._mode = mode
        self.text_ftr_size = len(hparams.filter_window_sizes) * hparams.num_filters

        self.embedding = model_utils.init_word_embedding(self._hparams, self._mode)
        with tf.variable_scope("cnn", dtype=tf.float32):
            self.query_ftrs = self.apply_cnn_on_query() if query is not None else None
            self.usr_ftrs = self.apply_cnn_on_usr() if usr_fields is not None else None
            self.doc_ftrs = self.apply_cnn_on_doc()

    def apply_cnn_on_query(self):
        """
        Apply cnn on queries

        :return Tensor  Query features. Shape=[batch_size, num_filters*len(filter_window_size)]
        """
        hparams = self._hparams
        query_emb = tf.nn.embedding_lookup(self.embedding, self._query)  # shape=[batch_size, sent_length, num_units]
        query_length = tf.reduce_sum(tf.cast(tf.not_equal(self._query, hparams.pad_id), dtype=tf.int32), axis=-1)
        query_ftrs = apply_cnn_on_text(query_emb,
                                       hparams.filter_window_sizes,
                                       hparams.num_filters,
                                       hparams.num_units,
                                       'query_cnn_',
                                       query_length,
                                       set_empty_zero=hparams.explicit_empty)
        return query_ftrs

    def apply_cnn_on_usr(self):
        """
        Apply cnn on user fields

        :return Tensor  User features. Shape=[batch_size, num_usr_fields, num_filters*len(filter_window_size)]
        """
        if type(self._usr_fields) is not list:
            return self._usr_fields

        hparams = self._hparams
        usr_ftrs = []
        for i, usr_field in enumerate(self._usr_fields):
            usr_emb = tf.nn.embedding_lookup(self.embedding, usr_field)  # shape=[batch_size, sent_length, num_units]
            usr_length = tf.reduce_sum(tf.cast(tf.not_equal(usr_field, hparams.pad_id), dtype=tf.int32), axis=-1)
            # shape=[batch_size, num_filters*len(filter_window_sizes)]
            usr_field_ftrs = apply_cnn_on_text(usr_emb,
                                               hparams.filter_window_sizes,
                                               hparams.num_filters,
                                               hparams.num_units,
                                               'usr_cnn_{}'.format(i),
                                               usr_length,
                                               set_empty_zero=hparams.explicit_empty)
            usr_ftrs.append(usr_field_ftrs)
        usr_ftrs = tf.stack(usr_ftrs, axis=1)  # shape=[batch_size, num_usr_fields, num_filters*len(filter_window_size)]
        return usr_ftrs

    def apply_cnn_on_doc(self):
        """Apply cnn on documents

        :return Tensor  Document features. Shape=[batch_size, max_group_size, num_doc_fields, num_ftrs]
        """

        # doc_fields should be the doc embeddings
        if type(self._doc_fields) is not list:
            return self._doc_fields

        hparams = self._hparams
        doc_ftrs = []
        for i, doc_field in enumerate(self._doc_fields):
            doc_emb = tf.nn.embedding_lookup(self.embedding,
                                             doc_field)  # shape=[batch_size, group_size, sent_length, num_units]

            # reshape docs to merge batch_size and group_size
            doc_emb_shape = tf.shape(doc_emb)
            doc_emb_reshape = tf.reshape(
                doc_emb, shape=[doc_emb_shape[0] * doc_emb_shape[1], doc_emb_shape[2], doc_emb_shape[3]])
            doc_length = tf.reduce_sum(tf.cast(tf.not_equal(doc_field, hparams.pad_id), dtype=tf.int32), axis=-1)
            doc_length = tf.reshape(doc_length, shape=[doc_emb_shape[0] * doc_emb_shape[1]])
            # apply cnn
            doc_field_ftrs = apply_cnn_on_text(doc_emb_reshape,
                                               hparams.filter_window_sizes,
                                               hparams.num_filters,
                                               hparams.num_units,
                                               'doc_cnn_' + str(i) + '_',
                                               doc_length,
                                               set_empty_zero=hparams.explicit_empty)
            # restore batch_size and group_size
            doc_field_ftrs = tf.reshape(doc_field_ftrs, shape=[
                doc_emb_shape[0], doc_emb_shape[1], self.text_ftr_size])
            doc_ftrs.append(doc_field_ftrs)
        doc_ftrs = tf.stack(doc_ftrs, axis=2)  # shape=[batch_size, max_group_size, num_doc_fields, num_ftrs]
        return doc_ftrs


def apply_cnn_on_text(text_emb,
                      filter_window_sizes,
                      num_filters,
                      num_units,
                      name_prefix,
                      text_length=None,
                      set_empty_zero=False):
    """
    For any text, apply several cnn with the same window size.
    The final embedding is concatenated from the cnn outputs.

    :return Tensor  Text features. Shape=[batch_size, num_filters*len(filter_window_sizes)]
    """
    text_ftrs = []
    for i in filter_window_sizes:
        t_conv = tf.layers.conv2d(
            inputs=tf.expand_dims(text_emb, 3),
            filters=num_filters,
            kernel_size=[i, num_units],
            activation=tf.nn.relu,
            padding='valid',
            name=name_prefix + str(i)
        )  # shape = [batch_size, q_len-filter_size+1, 1, num_filters]
        if text_length is not None:
            # Make sure the convolution results is consistent
            mask_length = tf.maximum(text_length - i + 1, 0)
            mask = tf.sequence_mask(mask_length, maxlen=tf.shape(text_emb)[1] - i + 1, dtype=tf.float32)
            mask = tf.expand_dims(tf.expand_dims(mask, axis=2), axis=2)
            t_conv *= mask

        t_pool = tf.reduce_max(t_conv, axis=1)  # shape = [batch_size, 1, num_filters]
        t_pool = tf.squeeze(t_pool, axis=1)  # shape = [batch_size, num_filters]

        # set the cnn output of empty strings as 0
        if set_empty_zero:
            not_empty_str = tf.cast(tf.not_equal(text_length, max(filter_window_sizes) * 2 - 2), dtype=tf.float32)
            t_pool *= tf.expand_dims(not_empty_str, axis=1)

        text_ftrs.append(t_pool)
    text_ftrs = tf.concat(text_ftrs, 1)  # shape = [batch_size, num_filters*len(filter_window_sizes)]
    return text_ftrs



# Copyright (c) LinkedIn Corporation. All rights reserved. Licensed under the BSD-2 Clause license.
# See LICENSE in the project root for license information.
import setuptools
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

requires = ['numpy<1.17',
            'smart-arg==0.4',
            'tensorflow==1.14.0',
            'tensorflow_ranking==0.1.4']
setuptools.setup(
    name='detext-nodep',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=["Programming Language :: Python :: 3.7",
                 "Intended Audience :: Science/Research",
                 "Intended Audience :: Developers",
                 "License :: OSI Approved"],
    license='BSD-2-CLAUSE',
    # DO NOT CHANGE: version should be incremented by bump2version when releasing. See pypi_release.sh
    version='2.1.1',
    package_dir={'': 'src'},
    packages=setuptools.find_packages('src'),
    include_package_data=True,
    install_requires=[],
    tests_require=[
        'pytest',
    ])



"""Tests for run_detext."""
import os
import shutil
import tensorflow as tf
import sys

from detext.run_detext import main, run_detext, DetextArg

root_dir = os.path.join(os.path.dirname(__file__), "resources")
out_dir = os.path.join(root_dir, "output")
data = os.path.join(root_dir, "train", "dataset", "tfrecord")
multitask_data = os.path.join(root_dir, "train", "multitask", "tfrecord")
vocab = os.path.join(root_dir, "vocab.txt")
bert_config = os.path.join(root_dir, "bert_config.json")


class TestModel(tf.test.TestCase):
    base_args = ["--all_metrics", "precision@1", "ndcg@10",
                 "--emb_sim_func", "inner", "concat", "diff",
                 "--elem_rescale", "True",
                 "--explicit_empty", "False",
                 "--init_weight", "0.1",
                 "--lambda_metric", "None",
                 "--learning_rate", "0.002",
                 "--ltr_loss_fn", "softmax",
                 "--max_gradient_norm", "1.0",
                 "--max_len", "16",
                 "--min_len", "3",
                 "--num_filters", "50",
                 "--num_train_steps", "4",
                 "--num_units", "4",
                 "--optimizer", "bert_adam",
                 "--pmetric", "ndcg@10",
                 "--steps_per_stats", "1",
                 "--steps_per_eval", "2",
                 "--train_batch_size", "2",
                 "--test_batch_size", "2",
                 "--use_deep", "True",
                 "--vocab_file", vocab,
                 "--vocab_file_for_id_ftr", vocab,
                 "--resume_training", "False",
                 "--unknown_param", "unknown_value"]  # added unknown param to test arg parsing

    args = base_args + [
        "--feature_names", "label,query,doc_completedQuery,usr_headline,usr_skills,usr_currTitles,"
                           "usrId_currTitles,docId_completedQuery,wide_ftrs,weight",
        "--num_wide", "3",
        "--test_file", data,
        "--dev_file", data,
        "--train_file", data,
    ]

    multitask_args = base_args + [
        "--feature_names", "query,label,wide_ftrs,doc_field1,doc_field2,task_id",
        "--num_wide", "10",
        "--test_file", multitask_data,
        "--dev_file", multitask_data,
        "--train_file", multitask_data,
    ]

    def _cleanUp(self, dir):
        if os.path.exists(dir):
            shutil.rmtree(dir, ignore_errors=True)

    def test_run_detext_cnn(self):
        """
        This method test run_detext with CNN models
        """
        output = os.path.join(out_dir, "cnn_model")
        args = self.args + \
            ["--filter_window_size", "1", "2", "3",
             "--ftr_ext", "cnn",
             "--num_hidden", "10", "10", "5",
             "--out_dir", output]
        sys.argv[1:] = args
        main(sys.argv)
        self._cleanUp(output)

    def test_run_detext_lstm(self):
        """
        This method test run_detext with LSTM models
        """
        output = os.path.join(out_dir, "lstm")
        args = self.args + \
            ["--filter_window_size", "0",
             "--ftr_ext", "lstm",
             "--num_hidden", "10",
             "--out_dir", output]
        sys.argv[1:] = args
        main(sys.argv)
        self._cleanUp(output)

    def test_run_detext_bert(self):
        """
        This method test run_detext with BERT models
        """
        output = os.path.join(out_dir, "bert")
        argument = DetextArg(
            ftr_ext='bert', num_units=4, num_wide=3,
            ltr_loss_fn='softmax', emb_sim_func=['inner', 'concat', 'diff'], num_filters=50,
            bert_config_file=bert_config,
            use_bert_dropout=True, optimizer='bert_adam',
            learning_rate=0.002, num_train_steps=4,
            train_batch_size=2, test_batch_size=2,
            train_file=data,
            dev_file=data,
            test_file=data,
            out_dir=output, max_len=16,
            vocab_file=vocab,
            vocab_file_for_id_ftr=vocab,
            steps_per_stats=1, steps_per_eval=2, keep_checkpoint_max=5,
            feature_names=['label', 'query', 'doc_completedQuery', 'usr_headline', 'usr_skills', 'usr_currTitles',
                           'usrId_currTitles', 'docId_completedQuery', 'wide_ftrs', 'weight'],
            pmetric='ndcg@10', all_metrics=['precision@1', 'ndcg@10'])

        run_detext(argument)
        self._cleanUp(output)

    def test_run_detext_multitask(self):
        """
        This method test run_detext with multitasking models
        """
        output = os.path.join(out_dir, "multitask_model")
        args = self.multitask_args + \
            ["--filter_window_size", "3",
             "--ftr_ext", "cnn",
             "--num_hidden", "10",
             "--task_ids", "0,1",
             "--task_weights", "0.2,0.8",
             "--out_dir", output]
        sys.argv[1:] = args
        main(sys.argv)
        self._cleanUp(output)


def test_demo():
    from subprocess import run, PIPE
    from pathlib import Path
    completed_process = run(['sh', 'run_detext.sh'], stderr=PIPE, cwd=f'{Path(__file__).parent}/resources')
    assert completed_process.returncode == 0
    assert completed_process.stderr.endswith(b'metric/precision@1 = 1.0\n')



import tensorflow as tf
from tensorflow.python.tools.saved_model_cli import _show_inputs_outputs

PLACEHOLDER_SUFFIX = '_placeholder'


def generate_sub_model(original_saved_model_dir, export_model_dir, input_tensor_names, output_tensor_names):
    """
    Modify original saved model to generate sub model in saved model format based on given input & output tensor names.
    Use this function to create "sub-graphs" such that pre-computing some nodes in the original graph could be enabled.
    Nodes that are not on the path of the "sub-graph" will not be included in the new graph.
    :param original_saved_model_dir: original saved model (*.pb) dir.
    :param export_model_dir: new dir to save the generated model. If exists, the dir will be removed first.
    :param input_tensor_names: inputs to include in new model separated by comma.
    Ensure the tensor names are in the computation graph.
    Eg. 'doc_currTitles_placeholder:0,doc_headlines_placeholder:0'
    :param output_tensor_names: outputs to include in the new model.
    Ensure the tensor names are in the computation graph. Eg. 'doc_ftrs:0'
    :return:
    """
    # Remove existing dir
    print('************** Generating sub model ********************')
    print('Inputs: ' + input_tensor_names)
    print('Outputs: ' + output_tensor_names)
    if tf.gfile.Exists(export_model_dir):
        print('Removing old output dir' + export_model_dir)
        tf.gfile.DeleteRecursively(export_model_dir)

    # Get list of input and output tensors
    input_tensor_list = input_tensor_names.split(',')
    output_tensor_list = output_tensor_names.split(',')

    with tf.Session(graph=tf.Graph()) as sess:
        # Load saved model graph
        tf.saved_model.loader.load(sess, [tf.saved_model.tag_constants.SERVING], original_saved_model_dir)
        graph = tf.get_default_graph()

        # Get signature dict from list of input and output tensors
        inputs = get_signature_dict(graph, input_tensor_list)
        outputs = get_signature_dict(graph, output_tensor_list)

        # Save to converted_model_dir with updated inputs and outputs
        tf.saved_model.simple_save(
            sess,
            export_model_dir,
            inputs=inputs,
            outputs=outputs
        )
        print('New model saved! Location: ' + export_model_dir)
        print('New model signatures:')
        # Use saved_model_cli tool to show the model signatures
        _show_inputs_outputs(export_model_dir,
                             tf.saved_model.tag_constants.SERVING,
                             tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY,
                             indent=0)


def get_signature_dict(graph, tensor_names_list, placeholder_suffix=PLACEHOLDER_SUFFIX):
    """
    Generate dict (inputs or outputs) for saved_model. The key is the name of the input placeholders or output tensors,
    and value is the tensors. input_name is the substring of the tensor name removing all after '_placeholder' to match
    the naming style in default model.
    :param graph: graph used to get tensors.
    :param tensor_names_list: tensors to put into the returned dict.
    :return: a dictionary that contains the <name: tensor> mapping
    """
    signature_dict = {}
    for t in tensor_names_list:
        input_name = t.split(placeholder_suffix)[0]
        signature_dict[input_name] = graph.get_tensor_by_name(t)
    return signature_dict


def generate_serving_models_ps_l1(orig_model_dir):
    """
    Generate the sub models for people search l1 models.
    This will generate:
    1. A query model for computing query embeddings online.
    2. A doc model for computing document embeddings to pre-compute all doc embeddings offline.
    3. A sim-wide model for using similarity scores and wide features to compute final scores online.
    :param orig_model_dir: original model directory
    :return:
    """
    print('Original model signature:')
    # Use saved_model_cli tool to show the model signatures
    _show_inputs_outputs(orig_model_dir,
                         tf.saved_model.tag_constants.SERVING,
                         tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY,
                         indent=0)

    # Query model for online inferencing query embeddings
    input_tensors = 'query_placeholder:0'
    outputs_tensors = 'query_ftrs:0'
    generate_sub_model(orig_model_dir, orig_model_dir + '-query-model', input_tensors, outputs_tensors)

    # Document model for pre-computing document embeddings
    input_tensors = 'doc_currTitles_placeholder:0,doc_headlines_placeholder:0,doc_prevTitles_placeholder:0'
    outputs_tensors = 'doc_ftrs:0'
    generate_sub_model(orig_model_dir, orig_model_dir + '-doc-model', input_tensors, outputs_tensors)

    # Online inference model to compute final scores using similarity features and wide features
    input_tensors = 'sim_ftrs:0,wide_ftr_placeholder:0'
    outputs_tensors = 'final_scores:0'
    generate_sub_model(orig_model_dir, orig_model_dir + '-sim-wide-model', input_tensors, outputs_tensors)


if __name__ == '__main__':
    generate_serving_models_ps_l1('/tmp/psmodel/best_ndcg@10')



"""
Reference: https://github.com/bluecamel/best_checkpoint_copier
"""
import os
import tensorflow as tf


class Checkpoint(object):
    dir = None
    file = None
    score = None
    path = None

    def __init__(self, path, score):
        self.dir = os.path.dirname(path)
        self.file = os.path.basename(path)
        self.score = score
        self.path = path


class BestCheckpointCopier(tf.estimator.Exporter):
    """
    Copy the best checkpoints for best model evaluation on test after training, and export best savedmodels for serving.
    """
    checkpoints = None
    checkpoints_to_keep = None
    compare_fn = None
    name = None
    pmetric = None
    sort_key_fn = None
    sort_reverse = None

    def __init__(self,
                 serving_input_receiver_fn,
                 name='best_checkpoints',
                 checkpoints_to_keep=1,
                 exports_to_keep=1,
                 pmetric='Loss/total_loss',
                 compare_fn=lambda x, y: x.score < y.score,
                 sort_reverse=False,
                 eval_log_file=None):
        """
        Construct BestCheckpointCopier

        :param serving_input_receiver_fn: serving input function for savedmodel (online serving model).
        :param name: name (dir path) to save the checkpoints and savedmodels.
        The best checkpoints will be copied to ./<name> and the savedmodels are exported to ./export/<name>
        :param checkpoints_to_keep: number of best checkpoints to keep.
        :param exports_to_keep: number of best savedmodels to keep.
        :param pmetric: primary metric to use for determining whether best model is achieved,
        :param compare_fn: compare function to judge current model with previous best.
        :param sort_reverse: sort reverse logic for checking which checkpoints to keep/remove. The top
        <checkpoints_to_keep> checkpoints will be kept. Eg. if the checkpoint with a higher score is considered
        better (compare_fn=lambda x, y: x.score > y.score), then sort_reverse should be set as True.
        """
        self.serving_input_receiver_fn = serving_input_receiver_fn
        self.checkpoints = []
        self.checkpoints_to_keep = checkpoints_to_keep
        self._exports_to_keep = exports_to_keep
        self.compare_fn = compare_fn
        self.name = name
        self.pmetric = pmetric
        self.sort_reverse = sort_reverse
        self.eval_log_file = eval_log_file
        self._log("***** Evaluation on dev set during training *****")
        super().__init__()

    def _copyCheckpoint(self, checkpoint):
        """
        Copy the checkpoint.
        """
        desination_dir = self._destinationDir(checkpoint)
        tf.gfile.MkDir(desination_dir)

        for file in tf.gfile.Glob(r'{}*'.format(checkpoint.path)):
            tf.gfile.Copy(file, os.path.join(desination_dir, os.path.basename(file)), overwrite=True)

    def _writeCheckpointStatusFile(self, checkpoint):
        """
        Write to 'checkpoint' with current checkpoint info.
        """
        checkpointStatusFile = os.path.join(self._destinationDir(checkpoint), 'checkpoint')
        with tf.gfile.Open(checkpointStatusFile, 'w') as fout:
            fout.write("model_checkpoint_path: \"{}\"".format(checkpoint.file))

    def _destinationDir(self, checkpoint):
        """
        Best checkpoints directory.
        """
        return os.path.join(checkpoint.dir, self.name)

    def _keepCheckpoint(self, checkpoint):
        """
        Add checkpoint to keep.
        """
        self._log('keeping checkpoint {} with {} = {}'.format(checkpoint.file, self.pmetric, checkpoint.score))
        self.checkpoints.append(checkpoint)
        self.checkpoints = sorted(self.checkpoints, key=lambda x: x.score, reverse=self.sort_reverse)
        self._copyCheckpoint(checkpoint)
        # Write to 'checkpoint' file with best checkpoint info.
        self._writeCheckpointStatusFile(self.checkpoints[0])

    def _log(self, statement):
        """
        Formatting log.
        """
        print(statement)
        if self.eval_log_file is not None:
            with tf.gfile.Open(self.eval_log_file, 'a') as fout:
                fout.write(statement + '\n')

    def _pruneCheckpoints(self, checkpoint):
        """
        Keep top <checkpoints_to_keep> checkpoints.
        """
        destination_dir = self._destinationDir(checkpoint)
        for checkpoint in self.checkpoints[self.checkpoints_to_keep:]:
            self._log('removing old checkpoint {} with {} = {}'.format(checkpoint.file, self.pmetric, checkpoint.score))
            old_checkpoint_path = os.path.join(destination_dir, checkpoint.file)
            for file in tf.gfile.Glob(r'{}*'.format(old_checkpoint_path)):
                tf.gfile.Remove(file)
        self.checkpoints = self.checkpoints[0:self.checkpoints_to_keep]

    def _score(self, eval_result):
        """
        Get the score for metric to compare and add logging.
        :param eval_result: current eval result containing the step info and metrics
        :return:
        """
        self._log("## Step {}".format(eval_result.get('global_step', -1)))
        for metric in eval_result:
            if metric != self.pmetric and metric != 'global_step':
                cm = eval_result[metric]
                if not hasattr(cm, "__len__"):
                    self._log("{} : {}".format(metric, cm))
                else:
                    self._log("{} : ".format(cm))
                    self._log(str(cm))
        return float(eval_result[self.pmetric])

    def _shouldKeep(self, checkpoint):
        """
        Whether current checkpoint should be kept.
        """
        return len(self.checkpoints) < self.checkpoints_to_keep or self.compare_fn(checkpoint, self.checkpoints[-1])

    def export(self, estimator, export_path, checkpoint_path, eval_result, is_the_final_export):
        """
        Perform best checkpoint copy and best savedmodel export if current checkpoint outperforms previous ones.
        """
        export_result = None
        score = self._score(eval_result)
        checkpoint = Checkpoint(path=checkpoint_path, score=score)
        self._log('Checking checkpoint {}'.format(checkpoint.file))

        if self._shouldKeep(checkpoint):
            self._keepCheckpoint(checkpoint)
            self._pruneCheckpoints(checkpoint)
            export_result = estimator.export_saved_model(
                export_path,
                self.serving_input_receiver_fn,
                checkpoint_path=checkpoint_path)
            tf.estimator.BestExporter._garbage_collect_exports(self, export_path)
        else:
            self._log('skipping checkpoint {} with {} = {}'.format(checkpoint.file, self.pmetric, checkpoint.score))
        # print new line.
        self._log('')
        return export_result
